# -*- encoding:utf-8 -*-
from ultron.ump.factor.sell.atrn_stop import FactorAtrNStop
from ultron.ump.factor.sell.close_atrn_stop import FactorCloseAtrNStop
from ultron.ump.factor.sell.pre_atrn_stop import FactorPreAtrNStop
from ultron.ump.factor.sell.sell_nday import FactorSellNDay
from ultron.ump.factor.sell.sell_break import FactorSellBreak
from ultron.ump.factor.sell.sell_break import FactorSellXDBK
from ultron.ump.factor.buy.buy_break import FactorBuyBreak
from ultron.ump.factor.buy.buy_break import FactorBuyXDBK
from ultron.ump.factor.buy.buy_break import FactorBuyPutBreak
from ultron.ump.factor.buy.buy_break import FactorBuyPutXDBK
from ultron.ump.factor.buy.buy_demo import FactorSDBreak
from ultron.ump.factor.buy.buy_wd import FactorBuyXD

from ultron.ump.factor.buy.buy_demo import TwoDayBuy
from ultron.ump.factor.buy.buy_demo import WeekMonthBuy
from ultron.ump.factor.buy.buy_dm import DoubleMaBuy
from ultron.ump.factor.buy.buy_trend import UpDownTrend
from ultron.ump.factor.buy.buy_trend import DownUpTrend
from ultron.ump.factor.sell.sell_dm import DoubleMaSell

__all__ = [
    'FactorAtrNStop', 'FactorCloseAtrNStop', 'FactorPreAtrNStop',
    'FactorSellNDay', 'FactorSellBreak', 'FactorSellXDBK', 'FactorBuyBreak',
    'FactorBuyXDBK', 'FactorBuyPutBreak', 'FactorBuyPutXDBK', 'FactorSDBreak',
    'FactorBuyXD', 'TwoDayBuy', 'WeekMonthBuy', 'DoubleMaBuy', 'UpDownTrend',
    'DownUpTrend', 'DoubleMaSell'
]



# -*- coding: utf-8 -*-

cdef inline double sign(double x) nogil:
    if x > 0.:
        return 1.
    elif x < 0.:
        return -1.
    else:
        return 0.


cdef inline double consecutive_int_sum(double s, double e) nogil:
    return (e + s) * (e - s + 1.) / 2.



# -*- encoding:utf-8 -*-
"""回测结果评分模块"""
import numpy as np
import pandas as pd
from abc import ABCMeta, abstractmethod
from collections import namedtuple

from ultron.ump.core.fixes import six
from ultron.ump.metrics.metrics_base import MetricsBase


class ScoreTuple(
        namedtuple('ScoreTuple',
                   ('orders_pd', 'action_pd', 'capital', 'benchmark',
                    'buy_factors', 'sell_factors', 'stock_picks'))):
    """namedtuple扩展类，在GridSearch中包装回测参数及结果"""

    __slots__ = ()

    def __repr__(self):
        return "orders_pd:{}\naction_pd:{}\ncapital:{}\nbenchmark:{}\n" \
               "buy_factors:{}\nsell_factors:{}\nstock_picks:{}".format(
                self.orders_pd.info() if self.orders_pd is not None else 'zero order',
                self.action_pd.info() if self.action_pd is not None else 'zero action', self.capital, self.benchmark,
                self.buy_factors, self.sell_factors, self.stock_picks)


class BaseScorer(six.with_metaclass(ABCMeta, object)):
    """针对GridSearch的score_tuple_array进行评分抽象基类"""

    # noinspection PyUnresolvedReferences
    def __init__(self, score_tuple_array, *arg, **kwargs):
        """
        :param score_tuple_array: 承接GridSearch返回的ScoreTuple对象序列
        :param kwargs: 可选weights代表评分项权重， 可选metrics_class代表交易目标度量类
        """
        self.score_tuple_array = score_tuple_array
        self.score_dict = {}
        self.weights_cnt = -1
        # 设置度量项抽取函数select_score_func，度量名称columns_name, weights_cnt
        self._init_self_begin(arg, *arg, **kwargs)

        # 检测_init_self_begin中必须要子类设置的有没有设置ok
        if not hasattr(self, 'select_score_func'):
            raise RuntimeError('_init_self_begin must set select_score_func')
        if not hasattr(self, 'columns_name'):
            raise RuntimeError('_init_self_begin must set columns_name')

        # 如果有设置权重就分配权重否则等权重
        if 'weights' in kwargs and kwargs['weights'] is not None and len(
                kwargs['weights']) == self.weights_cnt:
            self.weights = kwargs['weights']
        else:
            self.weights = self.weights_cnt * [
                1. / self.weights_cnt,
            ]

        # metrics_class = kwargs.pop('metrics_class', AbuMetricsBase)
        if 'metrics_class' in kwargs and kwargs['metrics_class'] is not None \
                and issubclass(kwargs['metrics_class'], MetricsBase):
            self.metrics_class = kwargs['metrics_class']
        else:
            self.metrics_class = MetricsBase

        valid_score_tuple_array = []
        for ind, score_tuple in enumerate(self.score_tuple_array):
            # 一个一个的进行度量
            metrics = self.metrics_class(score_tuple.orders_pd,
                                         score_tuple.action_pd,
                                         score_tuple.capital,
                                         score_tuple.benchmark)
            if metrics.valid:
                metrics.fit_metrics()
                # 使用子类_init_self_begin中设置的select_score_func方法选取
                self.score_dict[ind] = self.select_score_func(metrics)
                valid_score_tuple_array.append(score_tuple)

        # 把筛选出来有交易结果的重新放到score_tuple_array中
        self.score_tuple_array = valid_score_tuple_array

        # 将score_dict转换DataFrame并且转制
        self.score_pd = pd.DataFrame(self.score_dict).T
        # 设置度量指标名称
        self.score_pd.columns = self.columns_name
        """
            一般的任务是将score_pd中需要反转的度量结果进行反转
            所以要在rank前做self._init_self_end
        """
        self._init_self_end(arg, *arg, **kwargs)

        # 分数每一项都由0-1
        score_ls = np.linspace(0, 1, self.score_pd.shape[0])
        for cn in self.columns_name:
            # 每一项的结果rank后填入对应项
            score = score_ls[(self.score_pd[cn].rank().values - 1).astype(int)]
            self.score_pd['score_' + cn] = score

        scores = self.score_pd.filter(regex='score_*')
        # 根据权重计算最后的得分
        self.score_pd['score'] = scores.apply(lambda s:
                                              (s * self.weights).sum(),
                                              axis=1)

    @abstractmethod
    def _init_self_begin(self, *arg, **kwargs):
        """子类需要实现，设置度量项抽取函数select_score_func，度量名称columns_name，weights_cnt"""
        pass

    @abstractmethod
    def _init_self_end(self, *arg, **kwargs):
        """子类需要实现，一般的任务是将score_pd中需要反转的度量结果进行反转"""
        pass

    def fit_score(self):
        """对度量结果按照score排序，返回排序后的score列"""
        self.score_pd.sort_values(by='score', inplace=True)
        return self.score_pd['score']

    def __call__(self):
        """call self.fit_score"""
        return self.fit_score()


class WrsmScorer(BaseScorer):

    def _init_self_begin(self, *arg, **kwargs):
        """胜率，策略收益，策略sharpe值，策略最大回撤组成select_score_func"""

        self.select_score_func = lambda metrics: [
            metrics.win_rate, metrics.algorithm_period_returns, metrics.
            algorithm_sharpe, metrics.max_drawdown
        ]
        self.columns_name = ['win_rate', 'returns', 'sharpe', 'max_drawdown']
        self.weights_cnt = len(self.columns_name)

    def _init_self_end(self, *arg, **kwargs):
        """
        _init_self_end这里一般的任务是将score_pd中需要反转的反转，默认是数据越大越好，有些是越小越好，
        类似make_scorer(xxx, greater_is_better=True)中的参数greater_is_better的作用：

                            sign = 1 if greater_is_better else -1

        WrsmScorer中max_drawdown虽然是越小越好，但由于本身值即是负数形式，所以不用反转数据
                win_rate	returns	sharpe	max_drawdown	score_win_rate	score_returns	score_sharpe	score_max_drawdown	score
        0	0.307087	0.256354	0.922678	-0.110116	0.189935	0.814123	0.601461	0.511769	0.529322
        1	0.307087	0.256354	0.922678	-0.110116	0.189935	0.814123	0.601461	0.511769	0.529322
        2	0.307087	0.256354	0.922678	-0.110116	0.189935	0.814123	0.601461	0.511769	0.529322
        3	0.307087	0.256354	0.922678	-0.110116	0.189935	0.814123	0.601461	0.511769	0.529322
        4	0.307087	0.256354	0.922678	-0.110116	0.189935	0.814123	0.601461	0.511769	0.529322
        """
        pass


class DemoScorer(BaseScorer):

    def _init_self_begin(self, *arg, **kwargs):
        """胜率，策略收益，手续费组成select_score_func"""

        self.select_score_func = lambda metrics: [
            metrics.win_rate, metrics.algorithm_period_returns, metrics.
            commission_sum
        ]
        self.columns_name = ['win_rate', 'returns', 'commission']
        self.weights_cnt = len(self.columns_name)

    def _init_self_end(self, *arg, **kwargs):
        """
        _init_self_end这里一般的任务是将score_pd中需要反转的反转，默认是数据越大越好，有些是越小越好，
        类似make_scorer(xxx, greater_is_better=True)中的参数greater_is_better的作用：

                            sign = 1 if greater_is_better else -1
        """
        pass


def make_scorer(score_tuple_array, sc_class, **kwargs):
    """
    score对外接口模块函数
    :param score_tuple_array: 承接GridSearch返回的ScoreTuple对象序列
    :param sc_class: 指定进行评分的具体评分类，AbuBaseScorer子类，非实例对象
    :param kwargs: AbuBaseScorer中init的参数，可选weights代表评分项权重， 可选metrics_class代表交易目标度量类
    :return: 通过AbuBaseScorer __call__ 调用sc_class.fit_score()，
             返回fit_score返回值，即self.score_pd.sort_values(by='score')['score']
    """
    return sc_class(score_tuple_array, **kwargs)()


# -*- encoding:utf-8 -*-

import sys
import weakref


class LazyFunc(object):
    """描述器类：作用在类中需要lazy的对象方法上"""

    def __init__(self, func):
        """
        外部使用eg：
            class BuyCallMixin(object):
                @LazyFunc
                def buy_type_str(self):
                    return "call"

                @LazyFunc
                def expect_direction(self):
                    return 1.0
        """
        self.func = func
        self.cache = weakref.WeakKeyDictionary()

    def __get__(self, instance, owner):
        """描述器__get__，使用weakref.WeakKeyDictionary将以实例化的instance加入缓存"""
        if instance is None:
            return self
        try:
            return self.cache[instance]
        except KeyError:
            ret = self.func(instance)
            self.cache[instance] = ret
            return ret

    def __set__(self, instance, value):
        """描述器__set__，raise AttributeError，即禁止外部set值"""
        raise AttributeError("LazyFunc set value!!!")

    def __delete__(self, instance):
        """描述器___delete__从weakref.WeakKeyDictionary cache中删除instance"""
        del self.cache[instance]


class LazyClsFunc(LazyFunc):
    """
        描述器类：
        作用在类中需要lazy的类方法上，实际上只是使用__get__(owner, owner)
        替换原始__get__(self, instance, owner)
    """

    def __get__(self, instance, owner):
        """描述器__get__，使用__get__(owner, owner)替换原始__get__(self, instance, owner)"""
        return super(LazyClsFunc, self).__get__(owner, owner)


def add_doc(func, doc):
    """Lazy add doc"""
    func.__doc__ = doc


def import_module(name):
    """Lazy impor _module"""
    __import__(name)
    return sys.modules[name]



# -*- encoding:utf-8 -*-

import os, pdb, itertools, copy, datetime, sys, collections

os.environ['ULTRON_DATA'] = 'keim'
sys.path.insert(0, os.path.abspath('../'))

import numpy as np
import pandas as pd

from ultron.env import *
from ultron.kdutils.kline import date_week_wave
from ultron.kdutils import scaler as scaler_util
from ultron.ump.indicator import nd
from ultron.optimize.grirdsearch.creater import MLCreater

enable_example_env()

btc = pd.read_csv(os.path.join('data', 'btc.csv'), index_col=0)
btc.index = pd.to_datetime(btc.index)
btc = btc.loc['2013-09-01':'2017-07-26']
btc.name = 'btc'

btc['big_wave'] = (btc.high - btc.low) / btc.pre_close > 0.055
btc['big_wave'] = btc['big_wave'].astype(int)
btc['big_wave'].value_counts()

btc_train_raw = btc[:-60]
btc_test_raw = btc[-60:]


def calc_ma(tc, ma):
    ma_key = 'ma{}'.format(ma)
    tc[ma_key] = nd.ma.calc_ma_from_prices(tc.close, ma, min_periods=1)


for ma in [5, 10, 21, 60]:
    calc_ma(btc_train_raw, ma)
    calc_ma(btc_test_raw, ma)


def btc_siblings_df(btc_raw):
    # 将所有交易日以3个为一组，切割成多个子df，即每一个子df中有3个交易日的交易数据
    btc_siblings = [
        btc_raw.iloc[sib_ind * 3:(sib_ind + 1) * 3, :]
        for sib_ind in np.arange(0, int(btc_raw.shape[0] / 3))
    ]

    btc_df = pd.DataFrame()
    for sib_btc in btc_siblings:
        # 使用数据标准化将连续3天交易日中的连续数值特征进行标准化操作
        sib_btc_scale = scaler_util.scaler_std(
            sib_btc.filter([
                'open', 'close', 'high', 'low', 'volume', 'pre_close', 'ma5',
                'ma10', 'ma21', 'ma60', 'atr21', 'atr14'
            ]))
        # 把标准化后的和big_wave，date_week连接起来
        sib_btc_scale = pd.concat(
            [sib_btc['big_wave'], sib_btc_scale, sib_btc['date_week']], axis=1)

        # 抽取第一天，第二天的大多数特征分别改名字以one，two为特征前缀，如：one_open，one_close，two_ma5，two_high.....
        a0 = sib_btc_scale.iloc[0].filter([
            'open', 'close', 'high', 'low', 'volume', 'pre_close', 'ma5',
            'ma10', 'ma21', 'ma60', 'atr21', 'atr14', 'date_week'
        ])
        a0.rename(index={
            'open': 'one_open',
            'close': 'one_close',
            'high': 'one_high',
            'low': 'one_low',
            'volume': 'one_volume',
            'pre_close': 'one_pre_close',
            'ma5': 'one_ma5',
            'ma10': 'one_ma10',
            'ma21': 'one_ma21',
            'ma60': 'one_ma60',
            'atr21': 'one_atr21',
            'atr14': 'one_atr14',
            'date_week': 'one_date_week'
        },
                  inplace=True)

        a1 = sib_btc_scale.iloc[1].filter([
            'open', 'close', 'high', 'low', 'volume', 'pre_close', 'ma5',
            'ma10', 'ma21', 'ma60', 'atr21', 'atr14', 'date_week'
        ])
        a1.rename(index={
            'open': 'two_open',
            'close': 'two_close',
            'high': 'two_high',
            'low': 'two_low',
            'volume': 'two_volume',
            'pre_close': 'two_pre_close',
            'ma5': 'two_ma5',
            'ma10': 'two_ma10',
            'ma21': 'two_ma21',
            'ma60': 'two_ma60',
            'atr21': 'two_atr21',
            'atr14': 'two_atr14',
            'date_week': 'two_date_week'
        },
                  inplace=True)
        # 第三天的特征只使用'open', 'low', 'pre_close', 'date_week'，该名前缀today，如today_open，today_date_week
        a2 = sib_btc_scale.iloc[2].filter(
            ['big_wave', 'open', 'low', 'pre_close', 'date_week'])
        a2.rename(index={
            'open': 'today_open',
            'low': 'today_low',
            'pre_close': 'today_pre_close',
            'date_week': 'today_date_week'
        },
                  inplace=True)
        # 将抽取改名字后的特征连接起来组合成为一条新数据，即3天的交易数据特征－>1条新的数据
        btc_df = btc_df.append(pd.concat([a0, a1, a2], axis=0),
                               ignore_index=True)
    return btc_df


btc_train0 = btc_siblings_df(btc_train_raw)
btc_train0.tail()

btc_train1 = btc_siblings_df(btc_train_raw[1:])
btc_train2 = btc_siblings_df(btc_train_raw[2:])

btc_train = pd.concat([btc_train0, btc_train1, btc_train2])
btc_train.index = np.arange(0, btc_train.shape[0])

dummies_one_week = pd.get_dummies(btc_train['one_date_week'],
                                  prefix='one_date_week')
dummies_two_week = pd.get_dummies(btc_train['two_date_week'],
                                  prefix='two_date_week')
dummies_today_week = pd.get_dummies(btc_train['today_date_week'],
                                    prefix='today_date_week')
btc_train.drop(['one_date_week', 'two_date_week', 'today_date_week'],
               inplace=True,
               axis=1)
btc_train = pd.concat(
    [btc_train, dummies_one_week, dummies_two_week, dummies_today_week],
    axis=1)

train_matrix = btc_train.values
y = train_matrix[:, 0]
x = train_matrix[:, 1:]

btc_ml = MLCreater()

btc_ml.random_forest_classifier_best(x, y)

pdb.set_trace()
print('-->')


# -*- coding: utf-8 -*-
import requests

version_url = 'https://ultronsandbox.oss-cn-hangzhou.aliyuncs.com/version/ultron.json'

__all__ = ['__version__']

__version__ = "1.0.7"


def get_version():
    res = requests.get(version_url).json()
    if res.get('code') != 200:
        return '', ''

    remote_version = res['data']['version']
    content = res['data']['content']

    return remote_version, content


def check_version():
    remote_version, content = get_version()
    if not remote_version or remote_version <= __version__:
        return
    print('Ultron 版本由%s升级到%s,升级内容%s,执行 pip install --uprgrade Finance-Ultron' %
          (__version__, remote_version, content))


check_version()


# -*- coding: utf-8 -*-

import copy
from copy import deepcopy
import six
from libc.math cimport sqrt
from libc.math cimport fabs
from libc.math cimport log
from libc.math cimport exp
from libc.math cimport acos
from libc.math cimport acosh
from libc.math cimport asin
from libc.math cimport asinh
from libc.math cimport ceil
from libc.math cimport floor
from libc.math cimport round
from libc.math cimport isnan
cimport cython
import numpy as np
cimport numpy as np
import pandas as pd
from ultron.sentry.Utilities.Asserts cimport pyFinAssert
from ultron.sentry.Math.MathConstants cimport NAN
from ultron.sentry.Math.udfs cimport sign
from ultron.sentry.Math.Distributions.NormalDistribution cimport InverseCumulativeNormal as InvNormImpl



cdef class IAccumulator(object):

    def __add__(self, right):
        return AddedValueHolder(self, right)

    def __sub__(self, right):
        return MinusedValueHolder(self, right)

    def __mul__(self, right):
        return MultipliedValueHolder(self, right)

    def __div__(self, right):
        return DividedValueHolder(self, right)

    # only work for python 3
    def __truediv__(self, right):
        return DividedValueHolder(self, right)

    def __richcmp__(self, right, int op):
        if op == 0:
            return LtOperatorValueHolder(self, right)
        elif op == 1:
            return LeOperatorValueHolder(self, right)
        elif op == 2:
            return EqOperatorValueHolder(self, right)
        elif op == 3:
            return NeOperatorValueHolder(self, right)
        elif op == 4:
            return GtOperatorValueHolder(self, right)
        elif op == 5:
            return GeOperatorValueHolder(self, right)

    def __rshift__(self, right):
        if isinstance(right, IAccumulator):
            return CompoundedValueHolder(self, right)

        try:
            return right(self)
        except TypeError:
            raise ValueError('{0} is not recognized as a valid operator'.format(right))

    def __neg__(self):
        return Negative(self)


cdef class Accumulator(IAccumulator):

    def __init__(self):
        self._isFull = False
        self._window = 0

    cpdef push(self, dict data):
        pass

    cpdef double result(self):
        pass

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef transform(self, data, str name=None, bint to_sort=False):

        cdef int i
        cdef int k
        cdef double[:, :] matrix_values
        cdef size_t n = len(data)
        cdef double[:] output_values = np.zeros(n)
        cdef list columns
        cdef size_t column_length
        cdef dict data_dict

        if to_sort:
            data.sort_index(inplace=True)

        if not name:
            name = 'transformed'

        data = data.select_dtypes([np.number])
        matrix_values = data.values.astype(float)
        columns = data.columns.tolist()
        column_length = len(columns)

        data_dict = {columns[k]: 0 for k in range(column_length)}

        for i in range(n):
            for k in range(column_length):
                data_dict[columns[k]] = matrix_values[i, k]
            self.push(data_dict)
            output_values[i] = self.result()

        return pd.Series(output_values, index=data.index, name=name)

    @property
    def value(self):
        return self.result()

    cpdef bint isFull(self):
        return self._isFull

    @property
    def window(self):
        return self._window

    @property
    def dependency(self):
        return self._dependency


cdef class Negative(Accumulator):

    def __init__(self, x):
        super(Negative, self).__init__()
        self._window = x.window
        self._isFull = False
        self._inner = copy.deepcopy(x)
        self._dependency = self._inner.dependency

    cpdef push(self, dict data):
        self._inner.push(data)
        self._isFull = self._isFull if self._isFull else self._inner.isFull()

    def __str__(self):
        return "-{0}".format(str(self._inner))

    cpdef double result(self):
        return -self._inner.result()


cdef class CombinedValueHolder(Accumulator):

    def __init__(self, left, right):
        super(CombinedValueHolder, self).__init__()
        self._left = build_holder(left)
        self._right = build_holder(right)
        self._dependency = list(set(self._left.dependency + self._right.dependency))
        self._window = max(self._left.window, self._right.window)
        self._isFull = False

    cpdef push(self, dict data):
        self._left.push(data)
        self._right.push(data)
        self._isFull = self._isFull or (self._left.isFull() and self._right.isFull())


cdef class AddedValueHolder(CombinedValueHolder):
    def __init__(self, left, right):
        super(AddedValueHolder, self).__init__(left, right)

    cpdef double result(self):
        cdef double res1 = self._left.result()
        cdef double res2 = self._right.result()
        return res1 + res2

    def __str__(self):
        return "{0} + {1}".format(str(self._left), str(self._right))


cdef class MinusedValueHolder(CombinedValueHolder):
    def __init__(self, left, right):
        super(MinusedValueHolder, self).__init__(left, right)

    cpdef double result(self):
        cdef double res1 = self._left.result()
        cdef double res2 = self._right.result()
        return res1 - res2

    def __str__(self):
        return "{0} - {1}".format(str(self._left), str(self._right))


cdef class MultipliedValueHolder(CombinedValueHolder):
    def __init__(self, left, right):
        super(MultipliedValueHolder, self).__init__(left, right)

    cpdef double result(self):
        cdef double res1 = self._left.result()
        cdef double res2 = self._right.result()
        return res1 * res2

    def __str__(self):

        if isinstance(self._left, (AddedValueHolder, MinusedValueHolder)):
            s1 = "({0})".format(str(self._left))
        else:
            s1 = "{0}".format(str(self._left))

        if isinstance(self._right,  (AddedValueHolder, MinusedValueHolder)):
            s2 = "({0})".format(str(self._right))
        else:
            s2 = "{0}".format(str(self._right))

        return "{0} \\times {1}".format(s1, s2)


cdef class DividedValueHolder(CombinedValueHolder):
    def __init__(self, left, right):
        super(DividedValueHolder, self).__init__(left, right)

    @cython.cdivision(True)
    cpdef double result(self):
        cdef double res1 = self._left.result()
        cdef double res2 = self._right.result()
        return res1 / res2

    def __str__(self):
        return "\\frac{{{0}}}{{{1}}}".format(str(self._left), str(self._right))


cdef class LtOperatorValueHolder(CombinedValueHolder):
    def __init__(self, left, right):
        super(LtOperatorValueHolder, self).__init__(left, right)

    cpdef double result(self):
        cdef double res1 = self._left.result()
        cdef double res2 = self._right.result()
        return res1 < res2

    def __str__(self):

        if isinstance(self._left, CombinedValueHolder):
            s1 = "({0})".format(str(self._left))
        else:
            s1 = "{0}".format(str(self._left))

        if isinstance(self._right,  CombinedValueHolder):
            s2 = "({0})".format(str(self._right))
        else:
            s2 = "{0}".format(str(self._right))

        return "{0} \lt {1}".format(s1, s2)


cdef class LeOperatorValueHolder(CombinedValueHolder):
    def __init__(self, left, right):
        super(LeOperatorValueHolder, self).__init__(left, right)

    cpdef double result(self):
        cdef double res1 = self._left.result()
        cdef double res2 = self._right.result()
        return res1 <= res2

    def __str__(self):

        if isinstance(self._left, CombinedValueHolder):
            s1 = "({0})".format(str(self._left))
        else:
            s1 = "{0}".format(str(self._left))

        if isinstance(self._right,  CombinedValueHolder):
            s2 = "({0})".format(str(self._right))
        else:
            s2 = "{0}".format(str(self._right))

        return "{0} \le {1}".format(s1, s2)


cdef class GtOperatorValueHolder(CombinedValueHolder):
    def __init__(self, left, right):
        super(GtOperatorValueHolder, self).__init__(left, right)

    cpdef double result(self):
        cdef double res1 = self._left.result()
        cdef double res2 = self._right.result()
        return res1 > res2

    def __str__(self):

        if isinstance(self._left, CombinedValueHolder):
            s1 = "({0})".format(str(self._left))
        else:
            s1 = "{0}".format(str(self._left))

        if isinstance(self._right,  CombinedValueHolder):
            s2 = "({0})".format(str(self._right))
        else:
            s2 = "{0}".format(str(self._right))

        return "{0} \gt {1}".format(s1, s2)


cdef class GeOperatorValueHolder(CombinedValueHolder):
    def __init__(self, left, right):
        super(GeOperatorValueHolder, self).__init__(left, right)

    cpdef double result(self):
        cdef double res1 = self._left.result()
        cdef double res2 = self._right.result()
        return res1 >= res2

    def __str__(self):

        if isinstance(self._left, CombinedValueHolder):
            s1 = "({0})".format(str(self._left))
        else:
            s1 = "{0}".format(str(self._left))

        if isinstance(self._right,  CombinedValueHolder):
            s2 = "({0})".format(str(self._right))
        else:
            s2 = "{0}".format(str(self._right))

        return "{0} \ge {1}".format(s1, s2)


cdef class EqOperatorValueHolder(CombinedValueHolder):
    def __init__(self, left, right):
        super(EqOperatorValueHolder, self).__init__(left, right)

    cpdef double result(self):
        cdef double res1 = self._left.result()
        cdef double res2 = self._right.result()
        return res1 == res2

    def __str__(self):

        if isinstance(self._left, CombinedValueHolder):
            s1 = "({0})".format(str(self._left))
        else:
            s1 = "{0}".format(str(self._left))

        if isinstance(self._right,  CombinedValueHolder):
            s2 = "({0})".format(str(self._right))
        else:
            s2 = "{0}".format(str(self._right))

        return "{0} = {1}".format(s1, s2)


cdef class NeOperatorValueHolder(CombinedValueHolder):
    def __init__(self, left, right):
        super(NeOperatorValueHolder, self).__init__(left, right)

    cpdef double result(self):
        cdef double res1 = self._left.result()
        cdef double res2 = self._right.result()
        return res1 != res2

    def __str__(self):

        if isinstance(self._left, CombinedValueHolder):
            s1 = "({0})".format(str(self._left))
        else:
            s1 = "{0}".format(str(self._left))

        if isinstance(self._right,  CombinedValueHolder):
            s2 = "({0})".format(str(self._right))
        else:
            s2 = "{0}".format(str(self._right))

        return "{0} \\neq {1}".format(s1, s2)


cdef class Identity(Accumulator):

    def __init__(self, value):
        self._dependency = []
        self._window = 0
        self._value = value
        self._isFull = True

    cpdef push(self, dict data):
        pass

    cpdef double result(self):
        return self._value

    def __str__(self):
        return str(self._value)


cdef class Current(Accumulator):

    def __init__(self, x='x'):
        """
        Current should only be used as a vanilla named value holder
        :param x:
        """
        super(Current, self).__init__()
        self._window = 0
        self._isFull = True
        self._dependency = [x]
        self._current = NAN

    cpdef push(self, dict data):
        try:
            self._current = data[self._dependency[0]]
        except KeyError:
            self._current = NAN

    def __str__(self):
        return "''\\text{{{0}}}''".format(str(self._dependency[0]))

    cpdef double result(self):
        return self._current


cdef class Latest(Accumulator):

    def __init__(self, x='x', current_value=NAN):
        """
        Latest should only be used as a vanilla named value holder
        :param x:
        :param current_value:
        """
        super(Latest, self).__init__()
        self._window = 0
        self._isFull = True
        self._latest = current_value
        self._dependency = [x]

    cpdef push(self, dict data):
        cdef double res
        try:
            res = data[self._dependency[0]]
        except KeyError:
            pass
        else:
            if not isnan(res):
                self._latest = res

    def __str__(self):
        return "''\\text{{{0}}}''".format(str(self._dependency[0]))

    cpdef double result(self):
        return self._latest


cdef bint isanumber(a):
    cdef bint bool_a = True
    try:
        float(repr(a))
    except:
        bool_a = False

    return bool_a


cpdef build_holder(name):
    if isinstance(name, Accumulator):
        return deepcopy(name)
    elif isinstance(name, six.string_types):
        return Current(name)
    elif isanumber(name):
        return Identity(float(name))
    elif hasattr(name, '__iter__'):
        return build_holder(name[0])
    else:
        raise ValueError("({0}) is not recognized as a valid accumuator type".format(type(name)))


cdef class CompoundedValueHolder(Accumulator):

    def __init__(self, left, right):
        super(CompoundedValueHolder, self).__init__()
        self._left = build_holder(left)
        self._right = build_holder(right)
        self._window = self._left.window + self._right.window
        self._dependency = deepcopy(self._left.dependency)

    cpdef push(self, dict data):
        cdef double left_value
        self._left.push(data)
        left_value = self._left.result()
        self._right.push({self._right.dependency[0]: left_value})
        self._isFull = self._isFull or (self._left.isFull() and self._right.isFull())

    cpdef double result(self):
        return self._right.result()


cdef class IIF(Accumulator):

    def __init__(self, cond, left, right):
        super(IIF, self).__init__()
        self._cond = build_holder(cond)
        self._left = build_holder(left)
        self._right = build_holder(right)
        self._dependency = list(set(self._cond.dependency + self._left.dependency + self._right.dependency))
        self._window = max(self._cond.window, self._left.window, self._right.window)
        self._isFull = False

    cpdef push(self, dict data):
        self._cond.push(data)
        self._left.push(data)
        self._right.push(data)
        self._isFull = self._isFull or (self._cond.isFull() and self._left.isFull() and self._right.isFull())

    cpdef double result(self):
        return self._left.result() if self._cond.result() else self._right.result()


cdef class BasicFunction(Accumulator):

    def __init__(self, x, orig_value=NAN):
        super(BasicFunction, self).__init__()
        self._inner = build_holder(x)
        self._window = self._inner.window
        self._isFull = False
        self._dependency = self._inner.dependency
        self._origValue = orig_value

    cpdef push(self, dict data):
        self._inner.push(data)
        self._origValue = self._inner.result()
        self._isFull = self._isFull or self._inner.isFull()


cdef class Exp(BasicFunction):
    def __init__(self, x, orig_value=NAN):
        super(Exp, self).__init__(x, orig_value)

    cpdef double result(self):
        return exp(self._origValue)

    def __str__(self):
        return "\\exp({0})".format(str(self._inner))


cdef class Log(BasicFunction):
    def __init__(self, x, orig_value=NAN):
        super(Log, self).__init__(x, orig_value)

    cpdef double result(self):
        return log(self._origValue)

    def __str__(self):
        return "\\ln({0})".format(str(self._inner))


cdef class Sqrt(BasicFunction):
    def __init__(self, x, orig_value=NAN):
        super(Sqrt, self).__init__(x, orig_value)

    cpdef double result(self):
        return sqrt(self._origValue)

    def __str__(self):
        return "\\sqrt{{{0}}}".format(str(self._inner))


# due to the fact that pow function is much slower than ** operator
cdef class Pow(BasicFunction):

    def __init__(self, x, n, orig_value=NAN):
        super(Pow, self).__init__(x, orig_value)
        self._n = n

    cpdef double result(self):
        return self._origValue ** self._n

    def __str__(self):
        return "{0} ^ {{{1}}}".format(str(self._inner), self._n)


cdef class Abs(BasicFunction):
    def __init__(self, x, orig_value=NAN):
        super(Abs, self).__init__(x, orig_value)

    cpdef double result(self):
        return fabs(self._origValue)

    def __str__(self):
        return "\\left| {0} \\right|".format(str(self._inner))


cdef class Sign(BasicFunction):
    def __init__(self, x, orig_value=NAN):
        super(Sign, self).__init__(x, orig_value)

    cpdef double result(self):
        return sign(self._origValue)

    def __str__(self):
        return "\\mathrm{{sign}}({0})".format(str(self._inner))


cdef class Acos(BasicFunction):
    def __init__(self, x, orig_value=NAN):
        super(Acos, self).__init__(x, orig_value)

    cpdef double result(self):
        return acos(self._origValue)

    def __str__(self):
        return "\\mathrm{{ACos}}({0})".format(str(self._inner))


cdef class Acosh(BasicFunction):
    def __init__(self, x, orig_value=NAN):
        super(Acosh, self).__init__(x, orig_value)

    cpdef double result(self):
        return acosh(self._origValue)

    def __str__(self):
        return "\\mathrm{{ACosh}}({0})".format(str(self._inner))

cdef class Asin(BasicFunction):
    def __init__(self, x, orig_value=NAN):
        super(Asin, self).__init__(x, orig_value)

    cpdef double result(self):
        return asin(self._origValue)

    def __str__(self):
        return "\\mathrm{{ASin}}({0})".format(str(self._inner))


cdef class Asinh(BasicFunction):
    def __init__(self, x, orig_value=NAN):
        super(Asinh, self).__init__(x, orig_value)

    cpdef double result(self):
        return asinh(self._origValue)

    def __str__(self):
        return "\\mathrm{{ASinh}}({0})".format(str(self._inner))


cdef class NormInv(BasicFunction):
    def __init__(self, x, orig_value=NAN, fullAcc=False):
        super(NormInv, self).__init__(x, orig_value)
        self._inv = InvNormImpl(0., 1., fullAcc)

    cpdef double result(self):
        return self._inv.inv(self._origValue)

    def __str__(self):
        return "\\mathrm{{NormInv}}({0}, fullAcc={1})".format(str(self._inner), self._inv.isFullAcc)


cdef class Ceil(BasicFunction):
    def __init__(self, x, orig_value=NAN):
        super(Ceil, self).__init__(x, orig_value)

    cpdef double result(self):
        return ceil(self._origValue)

    def __str__(self):
        return "\\mathrm{{Ceil}}({0})".format(str(self._inner))


cdef class Floor(BasicFunction):
    def __init__(self, x, orig_value=NAN):
        super(Floor, self).__init__(x, orig_value)

    cpdef double result(self):
        return floor(self._origValue)

    def __str__(self):
        return "\\mathrm{{Floor}}({0})".format(str(self._inner))


cdef class Round(BasicFunction):
    def __init__(self, x, orig_value=NAN):
        super(Round, self).__init__(x, orig_value)

    cpdef double result(self):
        return round(self._origValue)

    def __str__(self):
        return "\\mathrm{{Round}}({0})".format(str(self._inner))
    
cdef class Sigmoid(BasicFunction):
    def __init__(self, x, orig_value=NAN):
        super(Sigmoid, self).__init__(x, orig_value)
    
    cpdef double result(self):
        return 1 / (1 + exp(self._origValue))
    
    def __str__(self):
        return "\\mathrm{{Sigmoid}}({0})".format(str(self._inner))
    
cdef class Tanh(BasicFunction):
    def __init__(self, x, orig_value=NAN):
        super(Tanh, self).__init__(x, orig_value)
    
    cpdef double result(self):
        return (exp(self._origValue) - exp(-self._origValue)) / (exp(self._origValue) + exp(-self._origValue))


# -*- coding: utf-8 -*-
import abc

import numpy as np
import pandas as pd

class Base(metaclass=abc.ABCMeta):
    def __init__(self):
        self.current_pos = pd.DataFrame()
    
    @abc.abstractmethod
    def execute(self, data: pd.DataFrame) -> pd.DataFrame:
        pass

    @staticmethod
    def calc_turn_over(target_pos: pd.DataFrame, current_pos: pd.DataFrame) -> float:
        pos_merged = pd.merge(target_pos, current_pos, on=['code'], how='outer')
        pos_merged.fillna(0, inplace=True)
        turn_over = np.abs(pos_merged.weight_x - pos_merged.weight_y).sum()
        return turn_over
    
    def set_current(self, current_pos: pd.DataFrame):
        self.current_pos = current_pos.copy()

    def update(self, data_dict: dict):
        pass
    



# -*- coding: utf-8 -*-

cimport cython
from ultron.sentry.Math.MathConstants cimport NAN
from cpython.mem cimport PyMem_Malloc, PyMem_Free
from libc.string cimport memcpy


cdef class Deque:

    def __cinit__(self,
                  size_t window):
        self.window = window
        self.is_full = False
        self.con = <double*> PyMem_Malloc(window * sizeof(double))
        for i in range(window):
            self.con[i] = 0.
        self.start = 0
        self.count = 0

    def __dealloc__(self):
        PyMem_Free(self.con)

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef double dump(self, double value, double default=NAN):
        cdef size_t n = self.start
        cdef double* con = self.con
        cdef double popout

        if self.is_full:
            popout = con[n]
            con[n] = value
            self.start = (n + 1) % self.window
            return popout
        else:
            con[self.count] = value
            self.count += 1
            self.is_full = self.count == self.window
            return default

    cpdef void dumps(self, values):
        for v in values:
            self.dump(v)

    cdef inline size_t size(self):
        return self.count

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef size_t idx(self, double value):
        cdef size_t i
        for i in range(self.count):
            if value == self.con[i]:
                break
        else:
            i = -1
        if i < 0:
            return -1
        else:
            i = (i - self.start + self.window) % self.window
        return i

    cdef inline bint isFull(self):
        return self.is_full

    @cython.wraparound(False)
    @cython.boundscheck(False)
    cpdef double sum(self):
        cdef double x = self.con[0]
        cdef int i
        for i in range(1, self.count):
            x += self.con[i]
        return x

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    def __getitem__(self, size_t item):
        return self.con[(self.start + item) % self.window]

    def __richcmp__(Deque self, Deque other, int op):
        cdef bint flag = False
        cdef int i
        if op == 2:
           flag =  self.window == other.window \
                   and self.is_full == other.is_full \
                   and self.start == other.start

           if flag:
               for i in range(self.window):
                   if self.con[i] != other.con[i]:
                       return False
               return True
           else:
               return False

        elif op == 3:
            return not self.__richcmp__(other, 2)

    cdef void set_data(self, bytes data):
        memcpy(self.con, <char*>data, sizeof(double)*self.window)

    def __reduce__(self):
        cdef bytes data = <bytes>(<char *>self.con)[:sizeof(double)*self.window]
        return rebuild, (data, self.window, self.is_full, self.start, self.count)


cpdef object rebuild(bytes data, size_t window, bint is_full, size_t start, size_t count):
    c = Deque(window)
    c.set_data(data)
    c.is_full = is_full
    c.start = start
    c.count = count
    return c


# -*- encoding:utf-8 -*-
"""
    卖出择时示例因子： 较小利润值 < 买入后最大收益价格 - 今日价格 < 较大利润值 －> 止盈卖出
    只做为单边止盈因子使用，作为利润保护因子使用
"""

from ultron.ump.factor.sell.base import FactorSellBase, ESupportDirection
"""外部可通过如：fs.close.g_default_close_atr_n = 2.5来修改默认值"""
g_default_close_atr_n = 3


class FactorCloseAtrNStop(FactorSellBase):
    """示例利润保护因子(止盈)因子"""

    def _init_self(self, **kwargs):
        """kwargs中可选参数close_atr_n: 保护利润止赢倍数"""

        self.close_atr_n = g_default_close_atr_n
        if 'close_atr_n' in kwargs:
            # 设置保护利润止赢倍数
            self.close_atr_n = kwargs['close_atr_n']
            self.sell_type_extra = '{}:close_atr_n={}'.format(
                self.__class__.__name__, self.close_atr_n)

    def support_direction(self):
        """单日最大跌幅n倍atr(止损)因子支持两个方向"""
        return [
            ESupportDirection.DIRECTION_CAll.value,
            ESupportDirection.DIRECTION_PUT.value
        ]

    def fit_day(self, today, orders):
        """
        止盈event： 较小利润值 < 买入后最大收益价格 - 今日价格 < 较大利润值
        :param today: 当前驱动的交易日金融时间序列数据
        :param orders: 买入择时策略中生成的订单序列
        :return:
        """

        for order in orders:
            # 通过order中的买入日期计算金融时间序列kl_pd中的index
            mask_date = self.kl_pd['date'] == order.buy_date
            start_ind = int(self.kl_pd[mask_date]['key'].values)
            end_ind = self.today_ind + 1
            """
                从买入日子开始计算到今天得到买入后最大收盘价格作为max_close，
                注意如果是call找序列中的最大收盘价格，put找序列中的最小收盘价格
            """
            max_close = self.kl_pd.iloc[start_ind:end_ind, :].close.max() if order.buy_type_str == 'call' \
                else self.kl_pd.iloc[start_ind:end_ind, :].close.min()
            """
                max_close - order.buy_price * 方向 > today['atr21']：代表只针对有一定盈利的情况生效，即 > 较小利润值
                max_close - today.close * 方向 > today['atr21'] * self.close_atr_n：下跌了一定值止盈退出, 即 < 较大利润值
            """
            if (max_close - order.buy_price) * order.expect_direction > today['atr21'] \
                    and (max_close - today.close) * order.expect_direction > today['atr21'] * self.close_atr_n:
                # 由于使用了当天的close价格，所以明天才能卖出
                self.sell_tomorrow(order)


# -*- encoding:utf-8 -*-
"""
    日内滑点卖出示例实现：均价卖出
    最简单的回测卖出方式，优点简单，且回测高效，在回测交易
    数量足够多的前提下也能接近实盘
"""

import numpy as np
from ultron.ump.slippage.sell_base import SlippageSellBase, slippage_limit_down


class SlippageSellMean(SlippageSellBase):
    """示例日内滑点均价卖出类"""

    @slippage_limit_down
    def fit_price(self):
        """
        取当天交易日的最高最低均价做为决策价格
        :return: 最终决策的当前交易卖出价格
        """

        self.sell_price = np.mean(
            [self.kl_pd_sell['high'], self.kl_pd_sell['low']])
        return self.sell_price


# -*- encoding:utf-8 -*-
"""
    资金模块，不区分美元，人民币等类型，做美股交易默认当作美元，a股默认当作人民币
"""
import numpy as np
import pandas as pd
from ultron.kdutils import date as date_utils
from ultron.utilities.logger import kd_logger
from ultron.kdutils.progress import Progress
from ultron.ump.core.base import PickleStateMixin
from ultron.ump.trade.commission import Commission
from ultron.ump.trade.order import Order


class Capital(PickleStateMixin):
    """资金类"""

    def __init__(self, init_cash, benchmark, user_commission_dict=None):
        """
        :param init_cash: 初始资金值，注意这里不区分美元，人民币等类型，做美股交易默认当作美元，a股默认当作人民币，int
        :param benchmark: 资金回测时间标尺，做为资金类表格时间范围确定使用，AbuBenchmark实例对象
        :param user_commission_dict: dict对象，可自定义交易手续费计算方法，详情查看AbuCommission
        """
        self.read_cash = init_cash

        kl_pd = benchmark.kl_pd
        if kl_pd is None:
            # 要求基准必须有数据
            raise ValueError('CapitalClass init klPd is None')

        # 根据基准时间序列，制作相同的时序资金对象capital_pd(pd.DcataFrame对象)
        self.capital_pd = pd.DataFrame(
            {
                'cash_blance': np.NAN * kl_pd.shape[0],
                'stocks_blance': np.zeros(kl_pd.shape[0]),
                'atr21': kl_pd['atr21'],
                'date': kl_pd['date']
            },
            index=kl_pd.index)

        self.capital_pd['date'] = self.capital_pd['date'].astype(int)
        # cash_blance除了第一个其它都是nan
        self.capital_pd.loc[self.capital_pd.index[0],
                            'cash_blance'] = self.read_cash
        # 构造交易手续费对象AbuCommission，如果user自定义手续费计算方法，通过user_commission_dict传入
        self.commission = Commission(user_commission_dict)

    def __str__(self):
        """打印对象显示：capital_pd.info commission_df.info"""
        return 'capital_pd:\n{}\ncommission_pd:\n{}'.format(
            self.capital_pd.info(), self.commission.commission_df.info())

    __repr__ = __str__

    def __len__(self):
        """对象长度：时序资金对象capital_pd的行数，即self.capital_pd.shape[0]"""
        return self.capital_pd.shape[0]

    def init_k_line(self, a_symbol):
        """
        每一个交易对象在时序资金对象capital_pd上都添加对应的call keep（买涨持仓量），call worth（买涨总价值），
        put keep（买跌持仓量），put worth（买跌总价值）
        :param a_symbol: symbol str对象
        """

        # 买涨持仓量
        call_keep = '_call_keep'
        if self.capital_pd.columns.tolist().count(a_symbol + call_keep) == 0:
            self.capital_pd[a_symbol + call_keep] = np.NAN * \
                                                    self.capital_pd.shape[0]
        # 买跌持仓量
        put_keep = '_put_keep'
        if self.capital_pd.columns.tolist().count(a_symbol + put_keep) == 0:
            self.capital_pd[a_symbol + put_keep] = np.NAN * \
                                                   self.capital_pd.shape[0]
        # 买涨总价值
        call_worth = '_call_worth'
        if self.capital_pd.columns.tolist().count(a_symbol + call_worth) == 0:
            self.capital_pd[a_symbol + call_worth] = np.NAN * \
                                                     self.capital_pd.shape[0]

        # 买跌总价值
        put_worth = '_put_worth'
        if self.capital_pd.columns.tolist().count(a_symbol + put_worth) == 0:
            self.capital_pd[a_symbol + put_worth] = np.NAN * \
                                                    self.capital_pd.shape[0]

    def apply_init_kl(self, action_pd, show_progress):
        """
        根据回测交易在时序资金对象capital_pd上新建对应的call，put列
        :param action_pd: 回测交易行为对象，pd.DataFrame对象
        :param show_progress: 外部设置是否需要显示进度条
        """
        # 使用set筛选唯一的symbol交易序列
        symbols = set(action_pd.symbol)
        # 单进程进度条
        with Progress(len(symbols), 0, label='apply_init_kl...') as progress:
            for pos, symbol in enumerate(symbols):
                if show_progress:
                    progress.show(a_progress=pos + 1)
                # 迭代symbols，新建对应的call，put列
                self.init_k_line(symbol)

    def apply_k_line(self, a_k_day, kl_pd, buy_type_head):
        """
        在apply_kl中的do_apply_kl方法中时序资金对象capital进行apply的对应方法，
        即迭代金融时间序列的每一个交易日，根据持仓量计算每一个交易日的市场价值
        :param a_k_day: 每一个被迭代中的时间，即每一个交易日数据
        :param kl_pd: 正在被apply迭代的金融时间序列本体，pd.DataFrame对象
        :param buy_type_head: 代表交易类型，范围（_call，_put）
        :return:
        """
        if a_k_day[kl_pd.name + buy_type_head + '_keep'] > 0:
            kl = kl_pd[kl_pd['date'] == a_k_day['date']]
            if kl is None or kl.shape[0] == 0:
                # 前提是当前交易日有对应的持仓
                return

            # 今天的收盘价格
            td_close = kl['close'].values[0]
            if buy_type_head == '_put':
                # 针对buy put对价格进行转换，主要是针对put特殊处理
                today_key = kl.key.values[0]
                if today_key > 0:
                    yd_close = kl_pd.iloc[today_key - 1].close
                    # 如果是买跌，实时市场收益以昨天为基础进行计算，即＊－1进行方向计算，以收益来映射重新定义今天的收盘价格
                    td_close = (td_close - yd_close) * -1 + yd_close
            # 根据持仓量即处理后的今日收盘价格，进行今日价值计算
            self.capital_pd.loc[kl.index, [kl_pd.name + buy_type_head + '_worth']] \
                = np.round(td_close * a_k_day[kl_pd.name + buy_type_head + '_keep'], 3)

    def apply_kl(self, action_pd, kl_pd_manager, show_progress):
        """
        apply_action之后对实际成交的交易分别迭代更新时序资金对象capital_pd上每一个交易日的实时价值
        :param action_pd: 回测结果生成的交易行为构成的pd.DataFrame对象
        :param kl_pd_manager: 金融时间序列管理对象，AbuKLManager实例
        :param show_progress: 是否显示进度条
        """

        # 在apply_action之后形成deal列后，set出考虑资金下成交了的交易序列
        deal_symbols_set = set(action_pd[action_pd['deal'] == 1].symbol)

        def do_apply_kl(kl_pd, buy_type_head):
            """
            根据金融时间序列在时序资金对象capital_pd上进行call（买涨），put（买跌）的交易日实时价值更新
            :param kl_pd: 金融时间序列，pd.DataFrame对象
            :param buy_type_head: 代表交易类型，范围（_call，_put）
            """
            # cash_blance对na进行pad处理
            self.capital_pd['cash_blance'].fillna(method='pad', inplace=True)
            # symbol对应列持仓量对na进行处理
            self.capital_pd[kl_pd.name + buy_type_head + '_keep'].fillna(
                method='pad', inplace=True)
            self.capital_pd[kl_pd.name + buy_type_head + '_keep'].fillna(
                0, inplace=True)

            # 使用apply在axis＝1上，即每一个交易日上对持仓量及市场价值进行更新
            self.capital_pd.apply(self.apply_k_line,
                                  axis=1,
                                  args=(kl_pd, buy_type_head))

            # symbol对应列市场价值对na进行处理
            self.capital_pd[kl_pd.name + buy_type_head + '_worth'].fillna(
                method='pad', inplace=True)
            self.capital_pd[kl_pd.name + buy_type_head + '_worth'].fillna(
                0, inplace=True)

            # 纠错处理把keep=0但是worth被pad的进行二次修正
            fe_mask = (self.capital_pd[kl_pd.name + buy_type_head + '_keep']
                       == 0) & (self.capital_pd[kl_pd.name + buy_type_head +
                                                '_worth'] > 0)
            # 筛出需要纠错的index
            fe_index = self.capital_pd[fe_mask].index
            # 将需要纠错的对应index上市场价值进行归零处理
            cp_w = self.capital_pd[kl_pd.name + buy_type_head + '_worth']
            cp_w.loc[fe_index] = 0

        # 单进程进度条
        with Progress(len(deal_symbols_set), 0,
                      label='apply_kl...') as progress:
            for pos, deal_symbol in enumerate(deal_symbols_set):
                if show_progress:
                    progress.show(a_progress=pos + 1)
                # 从kl_pd_manager中获取对应的金融时间序列kl，每一个kl分别进行call（买涨），put（买跌）的交易日实时价值更新
                kl = kl_pd_manager.get_pick_time_kl_pd(deal_symbol)
                # 进行call（买涨）的交易日实时价值更新
                do_apply_kl(kl, '_call')
                # 进行put（买跌）的交易日实时价值更新
                do_apply_kl(kl, '_put')

    def apply_action(self, a_action, progress):
        """
        在回测结果生成的交易行为构成的pd.DataFrame对象上进行apply对应本方法，即
        将交易行为根据资金情况进行处理，处理手续费以及时序资金对象capital_pd上的
        数据更新
        :param a_action: 每一个被迭代中的action，即每一个交易行为
        :param progress: 进度条对象
        :return: 是否成交deal bool
        """
        # 区别买入行为和卖出行为
        is_buy = True if a_action['action'] == 'buy' else False
        # 从action数据构造Order对象
        order = Order()
        order.buy_symbol = a_action['symbol']
        order.buy_cnt = a_action['Cnt']
        if is_buy:
            # 如果是买单，sell_price = price2 ,详情阅读ABuTradeExecute中transform_action
            order.buy_price = a_action['Price']
            order.sell_price = a_action['Price2']
        else:
            # 如果是卖单，buy_price = price2 ,详情阅读ABuTradeExecute中transform_action
            order.sell_price = a_action['Price']
            order.buy_price = a_action['Price2']
        # 交易发生的时间
        order.buy_date = a_action['Date']
        order.sell_date = a_action['Date']
        # 交易的方向
        order.expect_direction = a_action['Direction']

        # 对买单和卖单分别进行处理，确定是否成交deal
        deal = self.buy_stock(order) if is_buy else self.sell_stock(order)

        if progress is not None:
            progress.show()

        return deal

    def buy_stock(self, a_order):
        """
        在apply_action中每笔交易进行处理，根据买单计算cost，在时序资金对象capital_pd上修改对应cash_blance，
        以及更新对应symbol上的持仓量
        :param a_order: 在apply_action中由action转换的AbuOrder对象
        :return: 是否成交deal bool
        """

        # 首先使用commission对象计算手续费
        with self.commission.buy_commission_func(a_order) as (buy_func,
                                                              commission_list):
            commission = buy_func(a_order.buy_cnt, a_order.buy_price)
            # 将上下文管理器中返回的commission_list中添加计算结果commission，内部根据list长度决定写入手续费记录pd.DataFrame
            commission_list.append(commission)
        # cost = 买单数量 ＊ 单位价格 ＋ 手续费
        order_cost = a_order.buy_cnt * a_order.buy_price + commission
        # 买单时间转换成pd时间日期对象
        time_ind = pd.to_datetime(date_utils.fmt_date(a_order.buy_date))
        # pd时间日期对象置换出对应的index number
        num_index = self.capital_pd.index.tolist().index(time_ind)

        # cash_blance初始化init中除了第一个其它都是nan
        cash_blance = self.capital_pd['cash_blance'].dropna()

        # 先截取cash_blance中 < time_ind的时间序列
        bl_ss = cash_blance[cash_blance.index <= time_ind]
        if bl_ss is None or bl_ss.shape[0] == 0:
            kd_logger.info('bl_ss.shape[0] == 0 ' + str(a_order.buy_date))
            return False
        # 截取的bl_ss中选中最后一个元素做为判定买入时刻的cash值
        cash = bl_ss.iloc[-1]
        # 判定买入时刻的cash值是否能够钱买入
        if cash >= order_cost and a_order.buy_cnt > 0:
            # 够的话，买入，先将cash － cost
            cash -= order_cost
            # 根据a_order.expect_direction确定是要更新call的持仓量还是put的持仓量
            buy_type_keep = '_call_keep' if a_order.expect_direction == 1.0 else '_put_keep'
            # 前提1: 资金时间序列中有这个a_order.buy_symbol + buy_type_keep列
            has_cond1 = self.capital_pd.columns.tolist().count(
                a_order.buy_symbol + buy_type_keep) > 0
            # 前提2: 对应这个列从单子买入日期index开始，在dropna后还有shape说明本就有持仓
            has_cond2 = self.capital_pd[
                a_order.buy_symbol +
                buy_type_keep].iloc[:num_index + 1].dropna().shape[0] > 0

            keep_cnt = 0
            if has_cond1 and has_cond2:
                # 前提1 + 前提2->本就有持仓, 拿到之前的持仓量
                keep_cnt = self.capital_pd[a_order.buy_symbol +
                                           buy_type_keep].iloc[:num_index +
                                                               1].dropna()[-1]

            keep_cnt += a_order.buy_cnt

            # TODO 这里迁移之前逻辑，删除了一些未迁移模块的逻辑，所以看起来多写了一个流程，需要重构逻辑
            # 将计算好的cash更新到资金时间序列中对应的位置
            self.capital_pd.loc[time_ind, ['cash_blance']] = np.round(cash, 3)
            # 资金时间序列中之后的cash_blance也对应开始减去order_cost
            self.capital_pd.loc[cash_blance[cash_blance.index > time_ind].index, ['cash_blance']] \
                -= order_cost
            # 在更新持仓量前，取出之前的数值
            org_cnt = self.capital_pd.loc[time_ind][a_order.buy_symbol +
                                                    buy_type_keep]
            # 更新持仓量
            self.capital_pd.loc[time_ind, [a_order.buy_symbol +
                                           buy_type_keep]] = keep_cnt
            if not np.isnan(org_cnt):
                # 对多个因子作用在同一个symbol上，且重叠了持股时间提前更新之后的持仓量
                keep_pos = self.capital_pd[a_order.buy_symbol +
                                           buy_type_keep].dropna()
                self.capital_pd.loc[keep_pos[keep_pos.index > time_ind].index, [a_order.buy_symbol + buy_type_keep]] \
                    += a_order.buy_cnt
            return True
        else:
            return False

    def sell_stock(self, a_order):
        """
        在apply_action中每笔交易进行处理，根据卖单计算cost，在时序资金对象capital_pd上修改对应cash_blance，
        以及更新对应symbol上的持仓量
        :param a_order: 在apply_action中由action转换的AbuOrder对象
        :return: 是否成交deal bool
        """

        # 卖单时间转换成pd时间日期对象
        time_ind = pd.to_datetime(date_utils.fmt_date(a_order.sell_date))
        # # pd时间日期对象置换出对应的index number
        num_index = self.capital_pd.index.tolist().index(time_ind)
        # 根据a_order.expect_direction确定是要更新call的持仓量还是put的持仓量
        buy_type_keep = '_call_keep' if a_order.expect_direction == 1.0 else '_put_keep'
        # 前提1: 资金时间序列中有这个a_order.buy_symbol + buy_type_keep列
        has_cond1 = self.capital_pd.columns.tolist().count(a_order.buy_symbol +
                                                           buy_type_keep) > 0
        # 前提2: 对应这个列从单子买入日期index开始，在dropna后还有shape说明本就有持仓
        has_cond2 = self.capital_pd[
            a_order.buy_symbol + buy_type_keep].iloc[:num_index +
                                                     1].dropna().shape[0] > 0

        if has_cond1 and has_cond2:
            # 有持仓, 拿到之前的持仓量
            keep_cnt = self.capital_pd[a_order.buy_symbol +
                                       buy_type_keep].iloc[:num_index +
                                                           1].dropna()[-1]
            sell_cnt = a_order.buy_cnt

            if keep_cnt < sell_cnt:
                # 忽略一个问题，就算是当时买时的这个单子没有成交，这里也试图卖出当时设想买入的股数
                sell_cnt = keep_cnt
            if sell_cnt == 0:
                # 有可能由于没买入的单子，造成没有成交
                return False
            # 更新对应持仓量
            keep_cnt -= sell_cnt
            # 将卖出价格转换成call，put都可计算收益的价格，不要进行计算公式合并，保留冗余，便于理解
            sell_earn_price = (a_order.sell_price - a_order.buy_price
                               ) * a_order.expect_direction + a_order.buy_price
            order_earn = sell_earn_price * sell_cnt

            # 使用commission对象计算手续费
            with self.commission.sell_commission_func(a_order) as (
                    sell_func, commission_list):
                commission = sell_func(sell_cnt, a_order.sell_price)
                # 将上下文管理器中返回的commission_list中添加计算结果commission，内部根据list长度决定写入手续费记录pd.DataFrame
                commission_list.append(commission)

            # cash_blance初始化init中除了第一个其它都是nan
            cash_blance = self.capital_pd['cash_blance'].dropna()
            # 截取 < time_indd的cash_blance中最后一个元素做为cash值
            cash = cash_blance[cash_blance.index <= time_ind].iloc[-1]
            cash += (order_earn - commission)
            # 将计算好的cash更新到资金时间序列中对应的位置
            self.capital_pd.loc[time_ind, ['cash_blance']] = np.round(cash, 3)

            # TODO 这里迁移之前逻辑，删除了一些未迁移模块的逻辑，所以看起来多写了一个流程，需要重构逻辑
            self.capital_pd.loc[cash_blance[cash_blance.index > time_ind].index, ['cash_blance']] \
                += (order_earn - commission)

            org_cnt = self.capital_pd.loc[time_ind][a_order.buy_symbol +
                                                    buy_type_keep]
            # 更新持仓量
            self.capital_pd.loc[time_ind, [a_order.buy_symbol +
                                           buy_type_keep]] = keep_cnt

            if not np.isnan(org_cnt):
                # 针对diff factor same stock的情况
                keep_pos = self.capital_pd[a_order.buy_symbol +
                                           buy_type_keep].dropna()
                self.capital_pd.loc[keep_pos[keep_pos.index > time_ind].index, [a_order.buy_symbol + buy_type_keep]] \
                    -= sell_cnt
            return True
        else:
            return False



# -*- coding: utf-8 -*-


import copy
import sys
import operator
import six
import numpy as np
cimport numpy as np
import pandas as pd
cimport cython
from libc.math cimport isnan
from ultron.sentry.Analysis.SeriesValues cimport SeriesValues
from ultron.sentry.Utilities.Tools import to_dict
from ultron.sentry.Math.Accumulators.StatefulAccumulators cimport Shift
from ultron.sentry.Math.Accumulators.StatefulAccumulators cimport Delta
from ultron.sentry.Math.Accumulators.IAccumulators cimport Current
from ultron.sentry.Math.Accumulators.IAccumulators cimport Latest
from ultron.sentry.Math.Accumulators.IAccumulators cimport isanumber
from ultron.sentry.Math.MathConstants cimport NAN

if sys.version_info > (3, 0, 0):
    div_attr = "truediv"
else:
    div_attr = "div"


cdef class SecurityValueHolder(object):

    def __init__(self):
        self._window = 0
        self._holderTemplate = None
        self.updated = 0
        self.cached = None
        self._innerHolders = {}

    @property
    def symbolList(self):
        return list(self._innerHolders.keys())

    @property
    def fields(self):
        return self._dependency

    @property
    def window(self):
        return self._window

    @property
    def value(self):
        return self.value_all()

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef value_all(self):

        cdef np.ndarray[double, ndim=1] values
        cdef Accumulator holder
        cdef size_t n
        cdef int i

        if self.updated:
            return SeriesValues(self.cached.values, self.cached.name_mapping)
        else:
            keys = sorted(self._innerHolders.keys())
            n = len(keys)
            values = np.zeros(n)
            for i, name in enumerate(keys):
                try:
                    holder = self._innerHolders[name]
                    values[i] = holder.result()
                except ArithmeticError:
                    values[i] = NAN
            self.cached = SeriesValues(values, index=keys)
            self.updated = 1
            return self.cached

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef SeriesValues value_by_names(self, list names):
        cdef Accumulator holder
        cdef np.ndarray[double, ndim=1] res
        cdef int i
        cdef size_t n

        if self.updated:
            return self.cached[names]
        else:
            n = len(names)
            res = np.zeros(n)
            for i, name in enumerate(names):
                try:
                    holder = self._innerHolders[name]
                    res[i] = holder.result()
                except ArithmeticError:
                    res[i] = NAN
            return SeriesValues(res, index=names)

    cpdef double value_by_name(self, name):
        cdef Accumulator holder
        if self.updated:
            return self.cached[name]
        else:
            try:
                holder = self._innerHolders[name]
                return holder.result()
            except ArithmeticError:
                return NAN

    @property
    def holders(self):
        return self._innerHolders

    def isFullByName(self, name):
        return self._innerHolders[name].isFull()

    @property
    def isFull(self):
        for name in self._innerHolders:
            if not self._innerHolders[name].isFull():
                return False
        return True

    def __getitem__(self, filter):
        if isinstance(filter, SecurityValueHolder):
            return FilteredSecurityValueHolder(self, filter)
        elif isinstance(filter, int):
            return SecurityShiftedValueHolder(filter, self)
        else:
            return self.value_all()[filter]

    def __add__(self, right):
        return SecurityAddedValueHolder(self, right)

    def __sub__(self, right):
        return SecuritySubbedValueHolder(self, right)

    def __mul__(self, right):
        return SecurityMultipliedValueHolder(self, right)

    def __div__(self, right):
        return SecurityDividedValueHolder(self, right)

    def __truediv__(self, right):
        return SecurityDividedValueHolder(self, right)

    def __richcmp__(self, right, int op):

        if op == 0:
            return SecurityLtOperatorValueHolder(self, right)
        elif op == 1:
            return SecurityLeOperatorValueHolder(self, right)
        elif op == 2:
            return SecurityEqOperatorValueHolder(self, right)
        elif op == 3:
            return SecurityNeOperatorValueHolder(self, right)
        elif op == 4:
            return SecurityGtOperatorValueHolder(self, right)
        elif op == 5:
            return SecurityGeOperatorValueHolder(self, right)

    def __neg__(self):
        return SecurityNegValueHolder(self)

    def __invert__(self):
        return SecurityInvertValueHolder(self)

    def __and__(self, right):
        return SecurityAndOperatorValueHolder(self, right)

    def __rand__(self, left):
        return SecurityAndOperatorValueHolder(left, self)

    def __or__(self, right):
        return SecurityOrOperatorValueHolder(self, right)

    def __ror__(self, left):
        return SecurityOrOperatorValueHolder(left, self)

    def __xor__(self, right):
        return SecurityXorValuedHolder(self, right)

    def __rxor__(self, left):
        return SecurityXorValuedHolder(left, self)

    cpdef shift(self, int n):
        return SecurityShiftedValueHolder(n, self)

    cpdef transform(self, data, str name=None, str category_field=None, bint dropna=True):

        cdef str f
        cdef int dummy_category
        cdef np.ndarray total_category
        cdef double[:, :] matrix_values
        cdef list columns
        cdef double[:] output_values
        cdef double[:] narr_view
        cdef int j
        cdef size_t start_count
        cdef size_t end_count
        cdef list split_category
        cdef list split_values
        cdef dict dict_data

        for f in self._dependency:
            if f not in data:
                raise ValueError('({0}) dependency is not in input data'.format(f))

        dummy_category = 0
        if not category_field:
            category_field = 'dummy'
            data[category_field] = 1
            dummy_category = 1
            total_index = list(range(len(data)))
        else:
            total_index = data.index

        if not name:
            name = 'transformed'

        total_category = data[category_field].values
        data = data.select_dtypes([np.number])
        matrix_values = data.values.astype(float)
        columns = data.columns.tolist()
        split_category, split_values = to_dict(total_index, total_category.tolist(), matrix_values, columns)

        output_values = np.zeros(len(data))

        start_count = 0
        if not dummy_category:
            for j, dict_data in enumerate(split_values):
                self.push(dict_data)
                end_count = start_count + len(dict_data)
                narr_view = self.value_by_names(split_category[j]).values.astype(float)
                output_values[start_count:end_count]  = narr_view
                start_count = end_count
        else:
            for j, dict_data in enumerate(split_values):
                self.push(dict_data)
                output_values[j] = float(self.value_by_name(split_category[j][0]))

        df = pd.DataFrame(np.array(output_values), index=data.index, columns=[name])
        if not dummy_category:
            df[category_field] = total_category

        if dropna:
            df.dropna(inplace=True)

        return df


cdef class SecuritySingleValueHolder(SecurityValueHolder):
    def __init__(self, window, holderType, x):
        super(SecuritySingleValueHolder, self).__init__()
        self._compHolder = build_holder(x)
        self._dependency = self._compHolder.fields
        self._window = window + self._compHolder.window
        self._holderTemplate = holderType(window=window, x=str(self._compHolder))
        self._innerHolders = {
            name: copy.deepcopy(self._holderTemplate) for name in self._compHolder.symbolList
            }

    cpdef push(self, dict data):
        cdef SeriesValues sec_values
        cdef Accumulator holder
        cdef str dummy_name
        self.updated = 0

        dummy_name = str(self._compHolder)
        self._compHolder.push(data)
        sec_values = self._compHolder.value_all()

        for name in sec_values.index():
            try:
                holder = self._innerHolders[name]
                holder.push({dummy_name: sec_values[name]})
            except KeyError:
                holder = copy.deepcopy(self._holderTemplate)
                holder.push({dummy_name:  sec_values[name]})
                self._innerHolders[name] = holder

    def __str__(self):
        return str(self._holderTemplate)


cdef class SecurityBinaryValueHolder(SecurityValueHolder):
    def __init__(self, window, holderType, x, y):
        super(SecurityBinaryValueHolder, self).__init__()
        self._compHolder1 = build_holder(x)
        self._compHolder2 = build_holder(y)
        self._dependency = list(set(self._compHolder1.fields + self._compHolder2.fields))
        self._window = window + max(self._compHolder1.window, self._compHolder2.window)
        self._holderTemplate = holderType(window=window, x=str(self._compHolder1), y=str(self._compHolder2))
        self._innerHolders = {
            name: copy.deepcopy(self._holderTemplate) for name in self._compHolder1.symbolList
            }

    cpdef push(self, dict data):

        cdef SeriesValues sec_values1
        cdef SeriesValues sec_values2
        cdef Accumulator holder
        cdef str dummy_name1
        cdef str dummy_name2
        self.updated = 0

        dummy_name1 = str(self._compHolder1)
        dummy_name2 = str(self._compHolder2)
        self._compHolder1.push(data)
        self._compHolder2.push(data)
        sec_values1 = self._compHolder1.value_all()
        sec_values2 = self._compHolder2.value_all()

        for name in sec_values1.index():
            try:
                holder = self._innerHolders[name]
                holder.push({dummy_name1: sec_values1[name], dummy_name2: sec_values2[name]})
            except KeyError:
                holder = copy.deepcopy(self._holderTemplate)
                holder.push({dummy_name1: sec_values1[name], dummy_name2: sec_values2[name]})
                self._innerHolders[name] = holder

    def __str__(self):
        return str(self._holderTemplate)


cdef class SecurityStatelessSingleValueHolder(SecurityValueHolder):

    def __init__(self, holderType, x, **kwargs):
        super(SecurityStatelessSingleValueHolder, self).__init__()
        self._compHolder = build_holder(x)
        self._holderTemplate = holderType(x=str(self._compHolder), **kwargs)
        self._innerHolders = {name: copy.deepcopy(self._holderTemplate) for name in self._compHolder.symbolList}
        self._dependency = self._compHolder.fields

    cpdef push(self, dict data):

        cdef SeriesValues sec_values
        cdef Accumulator holder
        cdef str dummy_name
        self.updated = 0

        dummy_name = str(self._compHolder)
        self._compHolder.push(data)
        sec_values = self._compHolder.value_all()

        for name in sec_values.index():
            try:
                holder = self._innerHolders[name]
                holder.push({dummy_name: sec_values[name]})
            except KeyError:
                holder = copy.deepcopy(self._holderTemplate)
                holder.push({dummy_name:  sec_values[name]})
                self._innerHolders[name] = holder

    def __str__(self):
        return str(self._holderTemplate)


cdef class FilteredSecurityValueHolder(SecurityValueHolder):

    def __init__(self, computer, filtering):
        super(FilteredSecurityValueHolder, self).__init__()
        self._filter = copy.deepcopy(filtering)
        self._computer = copy.deepcopy(computer)
        self._window = max(computer.window, filtering.window)
        self._dependency = list(set(self._computer.fields + self._filter.fields))
        self.updated = 0
        self.cached = None

    def isFullByName(self, name):
        return self._computer.isFullByName(name)

    @property
    def isFull(self):
        return self._computer.isFull

    @property
    def symbolList(self):
        return self._computer.symbolList

    @property
    def holders(self):
        return self._computer.holders

    cpdef value_all(self):
        cdef np.ndarray[np.uint8_t, ndim=1, cast=True] filter_value

        if self.updated:
            return self.cached
        else:
            filter_value = self._filter.value_all().values.astype(bool)
            self.cached = self._computer.value_all().mask(filter_value)
            self.updated = 1
            return self.cached

    cpdef double value_by_name(self, name):

        cdef double filter_value

        if self.updated:
            return self.cached[name]
        else:
            filter_value = self._filter.value_by_name(name)
            if filter_value:
                return self._computer.value_by_name(name)
            else:
                return NAN

    cpdef SeriesValues value_by_names(self, list names):

        cdef SeriesValues filter_value
        cdef SeriesValues orig_values

        if self.updated:
            return self.cached[names]
        else:
            filter_value = self._filter.value_by_names(names)
            orig_values = self._computer.value_by_names(names)
            return SeriesValues(np.where(filter_value.values, orig_values.values, NAN), filter_value.name_mapping)

    cpdef push(self, dict data):
        self._computer.push(data)
        self._filter.push(data)
        self.updated = 0

    def __str__(self):
        return "\\mathrm{{Filter}}({0}, {1})".format(str(self._filter), str(self._computer))


cdef class IdentitySecurityValueHolder(SecurityValueHolder):

    def __init__(self, value):
        super(IdentitySecurityValueHolder, self).__init__()
        self._value = value
        self._dependency = []
        self._symbols = set()
        self.updated = 0
        self.cached = None

    def __str__(self):
        return str(self._value)

    def isFullByName(self, name):
        return True

    @property
    def isFull(self):
        return True

    cpdef push(self, dict data):
        self._symbols = self._symbols.union(data.keys())
        self.updated = 0

    cpdef SeriesValues value_all(self):
        return SeriesValues({n: self._value for n in self._symbols})

    cpdef double value_by_name(self, name):
        return self._value

    cpdef SeriesValues value_by_names(self, list names):
        return SeriesValues({n: self._value for n in names})

    def __str__(self):
        return "\\mathrm{{Identity}}({0})".format(str(self._value))


cdef class SecurityConstArrayValueHolder(SecurityValueHolder):

    def __init__(self, values):
        super(SecurityConstArrayValueHolder, self).__init__()
        self._dependency = []

        if isinstance(values, SeriesValues):
            self._values = values
        else:
            self._values = SeriesValues(values)

    def isFullByName(self, name):
        return True

    @property
    def isFull(self):
        return True

    cpdef push(self, dict data):
        pass

    cpdef double value_by_name(self, name):
        if name in self._values:
            return self._values[name]
        else:
            return NAN

    cpdef SeriesValues value_by_names(self, list names):
        return self._values[names]

    cpdef SeriesValues value_all(self):
        return self._values


cdef class SecurityUnitoryValueHolder(SecurityValueHolder):

    def __init__(self, right, op):
        super(SecurityUnitoryValueHolder, self).__init__()
        self._right = copy.deepcopy(right)

        self._window = self._right.window
        self._dependency = copy.deepcopy(self._right.fields)
        self._op = op
        self.updated = 0
        self.cached = None

    def isFullByName(self, name):
        return self._right.isFullByName(name)

    @property
    def isFull(self):
        return self._right.isFull

    @property
    def symbolList(self):
        return self._right.symbolList

    cpdef push(self, dict data):
        self._right.push(data)
        self.updated = 0

    cpdef SeriesValues value_all(self):
        if self.updated:
            return self.cached
        else:
            self.cached = self._op(self._right.value_all())
            self.updated = 1
            return self.cached

    cpdef double value_by_name(self, name):
        if self.updated:
            return self.cached[name]
        else:
            return self._op(self._right.value_by_name(name))

    cpdef SeriesValues value_by_names(self, list names):
        if self.updated:
            return self.cached[names]
        else:
            return self._op(self._right.value_by_names(names))


cdef class SecurityNegValueHolder(SecurityUnitoryValueHolder):
    def __init__(self, right):
        super(SecurityNegValueHolder, self).__init__(
            right, operator.neg)

    def __str__(self):
        return "-{0}".format(str(self._right))


cdef class SecurityInvertValueHolder(SecurityUnitoryValueHolder):
    def __init__(self, right):
        super(SecurityInvertValueHolder, self).__init__(
            right, operator.invert)


cdef class SecurityCurrentValueHolder(SecurityValueHolder):
    def __init__(self, x):
        super(SecurityCurrentValueHolder, self).__init__()
        self._dependency = [x]
        self._symbol_values = {}
        self._holderTemplate = Current(x)

    @property
    def symbolList(self):
        return list(self._symbol_values.keys())

    cpdef push(self, dict data):
        cdef double value
        cdef dict data_pack
        field = self._dependency[0]
        self.updated = 0

        for name in data:
            data_pack = data[name]
            if field in data_pack:
                value = data_pack[field]
                self._symbol_values[name] = value

            if name not in self._symbol_values:
                self._symbol_values[name] = NAN

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef SeriesValues value_all(self):

        cdef np.ndarray values
        cdef size_t n
        cdef int i

        if self.updated:
            return SeriesValues(self.cached.values, self.cached.name_mapping)
        else:
            keys = sorted(self._symbol_values.keys())
            n = len(keys)
            values = np.zeros(n)
            for i, name in enumerate(keys):
                values[i] = self._symbol_values[name]
            self.cached = SeriesValues(values, index=keys)
            self.updated = 1
            return self.cached

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef SeriesValues value_by_names(self, list names):
        cdef Accumulator holder
        cdef np.ndarray res
        cdef int i
        cdef size_t n

        if self.updated:
            return self.cached[names]
        else:
            n = len(names)
            res = np.zeros(n)
            for i, name in enumerate(names):
                res[i] = self._symbol_values[name]
            return SeriesValues(res, index=names)

    cpdef double value_by_name(self, name):
        cdef Accumulator holder
        if self.updated:
            return self.cached[name]
        else:
            return self._symbol_values[name]

    def isFullByName(self, name):
        return True

    @property
    def isFull(self):
        return True

    def __str__(self):
        return str(self._holderTemplate)


cdef class SecurityLatestValueHolder(SecurityValueHolder):
    def __init__(self, x):
        super(SecurityLatestValueHolder, self).__init__()
        self._dependency = [x]
        self._symbol_values = {}
        self._holderTemplate = Latest(x)

    @property
    def symbolList(self):
        return list(self._symbol_values.keys())

    cpdef push(self, dict data):
        cdef double value
        cdef dict data_pack
        field = self._dependency[0]
        self.updated = 0

        for name in data:
            data_pack = data[name]
            if field in data_pack:
                value = data_pack[field]
                if not isnan(value):
                    self._symbol_values[name] = value

            if name not in self._symbol_values:
                self._symbol_values[name] = NAN

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef SeriesValues value_all(self):

        cdef np.ndarray values
        cdef size_t n
        cdef int i

        if self.updated:
            return SeriesValues(self.cached.values, self.cached.name_mapping)
        else:
            keys = sorted(self._symbol_values.keys())
            n = len(keys)
            values = np.zeros(n)
            for i, name in enumerate(keys):
                values[i] = self._symbol_values[name]
            self.cached = SeriesValues(values, index=keys)
            self.updated = 1
            return self.cached

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef SeriesValues value_by_names(self, list names):
        cdef Accumulator holder
        cdef np.ndarray res
        cdef int i
        cdef size_t n

        if self.updated:
            return self.cached[names]
        else:
            n = len(names)
            res = np.zeros(n)
            for i, name in enumerate(names):
                res[i] = self._symbol_values[name]
            return SeriesValues(res, index=names)

    cpdef double value_by_name(self, name):
        cdef Accumulator holder
        if self.updated:
            return self.cached[name]
        else:
            return self._symbol_values[name]

    def isFullByName(self, name):
        return True

    @property
    def isFull(self):
        return True

    def __str__(self):
        return str(self._holderTemplate)


cpdef SecurityValueHolder build_holder(name):
    if isinstance(name, SecurityValueHolder):
        return copy.deepcopy(name)
    elif isinstance(name, six.string_types):
        return SecurityCurrentValueHolder(name)
    elif isanumber(name):
        return IdentitySecurityValueHolder(float(name))
    elif isinstance(name, (list, tuple)) and len(name) == 1:
        return SecurityCurrentValueHolder(name[0])
    else:
        raise ValueError("{0} is not recognized as valid holder or name".format(name))


cdef class SecurityCombinedValueHolder(SecurityValueHolder):

    def __init__(self, left, right, op):
        super(SecurityCombinedValueHolder, self).__init__()
        self._left = build_holder(left)
        self._right = build_holder(right)
        self._window = max(self._left.window, self._right.window)
        self._dependency = list(set(self._left.fields + self._right.fields))
        self._op = op
        self.updated = 0
        self.cached = None

    def isFullByName(self, name):
        return self._left.isFullByName(name) and self._right.isFullByName(name)

    @property
    def isFull(self):
        return self._left.isFull and self._right.isFull

    @property
    def symbolList(self):
        return list(set(self._left.symbolList + self._right.symbolList))

    cpdef push(self, dict data):
        self._left.push(data)
        self._right.push(data)
        self.updated = 0

    cpdef SeriesValues value_all(self):
        if self.updated:
            return self.cached
        else:
            self.cached = self._op(self._left.value_all(), self._right.value_all())
            self.updated = 1
            return self.cached

    cpdef double value_by_name(self, name):
        if self.updated:
            return self.cached[name]
        else:
            return self._op(self._left.value_by_name(name), self._right.value_by_name(name))

    cpdef SeriesValues value_by_names(self, list names):
        if self.updated:
            return self.cached[names]
        else:
            return self._op(self._left.value_by_names(names), self._right.value_by_names(names))


cdef class SecurityXorValuedHolder(SecurityCombinedValueHolder):
    def __init__(self, left, right):
        super(SecurityXorValuedHolder, self).__init__(
            left, right, operator.xor)

    cpdef double value_by_name(self, name):
        if self.updated:
            return self.cached[name]
        else:
            return self._left.value_by_name(name), self._right.value_by_name(name)

    def __str__(self):
        return "({0}, {1})".format(str(self._left), str(self._right))


cdef class SecurityAddedValueHolder(SecurityCombinedValueHolder):
    def __init__(self, left, right):
        super(SecurityAddedValueHolder, self).__init__(
            left, right, operator.add)

    def __str__(self):
        return "{0} + {1}".format(str(self._left), str(self._right))


cdef class SecuritySubbedValueHolder(SecurityCombinedValueHolder):
    def __init__(self, left, right):
        super(SecuritySubbedValueHolder, self).__init__(
            left, right, operator.sub)

    def __str__(self):
        return "{0} - {1}".format(str(self._left), str(self._right))


cdef class SecurityMultipliedValueHolder(SecurityCombinedValueHolder):
    def __init__(self, left, right):
        super(SecurityMultipliedValueHolder, self).__init__(
            left, right, operator.mul)

    def __str__(self):
        return "{0} * {1}".format(str(self._left), str(self._right))

cdef class SecurityDividedValueHolder(SecurityCombinedValueHolder):
    def __init__(self, left, right):
        super(SecurityDividedValueHolder, self).__init__(
            left, right, getattr(operator, div_attr))

    def __str__(self):
        return "\\frac{{{0}}}{{{1}}}".format(str(self._left), str(self._right))


cdef class SecurityLtOperatorValueHolder(SecurityCombinedValueHolder):
    def __init__(self, left, right):
        super(SecurityLtOperatorValueHolder, self).__init__(
            left, right, operator.lt)

    def __str__(self):
        return "{0} \\lt {1}".format(str(self._left), str(self._right))


cdef class SecurityLeOperatorValueHolder(SecurityCombinedValueHolder):
    def __init__(self, left, right):
        super(SecurityLeOperatorValueHolder, self).__init__(
            left, right, operator.le)

    def __str__(self):
        return "{0} \\le {1}".format(str(self._left), str(self._right))


cdef class SecurityGtOperatorValueHolder(SecurityCombinedValueHolder):
    def __init__(self, left, right):
        super(SecurityGtOperatorValueHolder, self).__init__(
            left, right, operator.gt)

    def __str__(self):
        return "{0} \\gt {1}".format(str(self._left), str(self._right))


cdef class SecurityGeOperatorValueHolder(SecurityCombinedValueHolder):
    def __init__(self, left, right):
        super(SecurityGeOperatorValueHolder, self).__init__(
            left, right, operator.ge)

    def __str__(self):
        return "{0} \\ge {1}".format(str(self._left), str(self._right))


cdef class SecurityEqOperatorValueHolder(SecurityCombinedValueHolder):
    def __init__(self, left, right):
        super(SecurityEqOperatorValueHolder, self).__init__(
            left, right, operator.eq)

    def __str__(self):
        return "{0} == {1}".format(str(self._left), str(self._right))


cdef class SecurityNeOperatorValueHolder(SecurityCombinedValueHolder):
    def __init__(self, left, right):
        super(SecurityNeOperatorValueHolder, self).__init__(
            left, right, operator.ne)

    def __str__(self):
        return "{0} \\neq {1}".format(str(self._left), str(self._right))


cdef class SecurityAndOperatorValueHolder(SecurityCombinedValueHolder):
    def __init__(self, left, right):
        super(SecurityAndOperatorValueHolder, self).__init__(
            left, right, operator.__and__)

    def __str__(self):
        return "{0} \\& {1}".format(str(self._left), str(self._right))


cdef class SecurityOrOperatorValueHolder(SecurityCombinedValueHolder):
    def __init__(self, left, right):
        super(SecurityOrOperatorValueHolder, self).__init__(
            left, right, operator.__or__)

    def __str__(self):
        return "{0} | {1}".format(str(self._left), str(self._right))


cdef class SecurityShiftedValueHolder(SecuritySingleValueHolder):

    def __init__(self, window, x):
        super(SecurityShiftedValueHolder, self).__init__(window, Shift, x)

    def __str__(self):
        return "\\mathrm{{Shift}}({0}, {1})".format(str(self._compHolder), self._holderTemplate.lag())


cdef class SecurityDeltaValueHolder(SecuritySingleValueHolder):

    def __init__(self, window, x):
        super(SecurityDeltaValueHolder, self).__init__(window, Delta, x)


cdef class SecurityIIFValueHolder(SecurityValueHolder):

    def __init__(self, flag, left, right):
        super(SecurityIIFValueHolder, self).__init__()
        self._flag = build_holder(flag)
        self._left = build_holder(left)
        self._right = build_holder(right)

        self._window = max(self._flag.window, self._left.window, self._right.window)
        self._dependency = list(set(self._flag.fields + self._left.fields + self._right.fields))
        self.updated = 0
        self.cached = None

    def isFullByName(self, name):
        return self._flag.isFullByName(name) and self._left.isFullByName(name) and self._right.isFullByName(name)

    @property
    def isFull(self):
        return self._flag.isFull and self._left.isFull and self._right.isFull

    @property
    def symbolList(self):
        return self._flag.symbolList

    cpdef push(self, dict data):
        self._flag.push(data)
        self._left.push(data)
        self._right.push(data)
        self.updated = 0

    cpdef SeriesValues value_all(self):

        cdef SeriesValues flag_value
        cdef np.ndarray[double, ndim=1] left_value
        cdef np.ndarray[double, ndim=1] right_value

        if self.updated:
            return self.cached
        else:
            flag_value = self._flag.value_all()

            left_value = self._left.value_all().values
            right_value = self._right.value_all().values

            self.cached = SeriesValues(np.where(flag_value.values,
                                                left_value,
                                                right_value),
                                       flag_value.name_mapping)
            self.updated = 1
            return self.cached

    cpdef double value_by_name(self, name):
        if self.updated:
            return self.cached[name]
        else:
            if self._flag.value_by_name(name):
                return self._left.value_by_name(name)
            else:
                return self._right.value_by_name(name)

    cpdef SeriesValues value_by_names(self, list names):

        cdef SeriesValues flag_value
        cdef np.ndarray[double, ndim=1] left_value
        cdef np.ndarray[double, ndim=1] right_value

        if self.updated:
            return self.cached[names]
        else:
            flag_value = self._flag.value_by_names(names)

            left_value = self._left.value_by_names(names).values
            right_value = self._right.value_by_names(names).values

            return SeriesValues(np.where(flag_value.values,
                                         left_value,
                                         right_value),
                                flag_value.name_mapping)

    def __str__(self):
        return "\\mathrm{{IIF}}({0}, {1}, {2})".format(str(self._flag), str(self._left), str(self._right))


# -*- encoding:utf-8 -*-
"""借鉴sklearn GridSearch，针对买入因子，卖出因子，选股因子最合进行最优寻找分析"""
import operator
from collections import Mapping
from functools import reduce
from itertools import product, chain
import numpy as np
from ultron.ump.core import env
from ultron.ump.core.fixes import partial
from ultron.utilities.logger import kd_logger
from ultron.kdutils.parallel import delayed, Parallel
from ultron.ump.core.process import add_process_env_sig, EnvProcess
from ultron.ump.metrics.metrics_base import MetricsBase
from ultron.ump.metrics.score import make_scorer, ScoreTuple, WrsmScorer
from ultron.ump.market.benchmark import Benchmark
from ultron.ump.trade.kl_manager import KLManager
from ultron.ump.trade.capital import Capital
from ultron.ump.alpha.pick_time_master import PickTimeMaster
from ultron.kdutils.progress import MulPidProgress
from ultron.ump.market.market import split_k_market


class ParameterGrid(object):
    """参数进行product辅助生成类"""

    def __init__(self, param_grid):
        if isinstance(param_grid, Mapping):
            param_grid = [param_grid]
        self.param_grid = param_grid

    def __iter__(self):
        """迭代参数组合实现"""
        for p in self.param_grid:
            items = sorted(p.items())
            if not items:
                yield {}
            else:
                keys, values = zip(*items)
                for v in product(*values):
                    params = dict(zip(keys, v))
                    yield params

    def __len__(self):
        """参数组合长度实现"""
        product_mul = partial(reduce, operator.mul)
        return sum(
            product_mul(len(v) for v in p.values()) if p else 1
            for p in self.param_grid)

    def __getitem__(self, ind):
        """通过index方式获取某个参数组合实现"""
        for sub_grid in self.param_grid:
            if not sub_grid:
                if ind == 0:
                    return {}
                else:
                    ind -= 1
                    continue
            keys, values_lists = zip(*sorted(sub_grid.items())[::-1])
            sizes = [len(v_list) for v_list in values_lists]
            total = np.product(sizes)

            if ind >= total:
                ind -= total
            else:
                out = {}
                for key, v_list, n in zip(keys, values_lists, sizes):
                    ind, offset = divmod(ind, n)
                    out[key] = v_list[offset]
                return out

        raise IndexError('ParameterGrid index out of range')


def _check_param_grid(param_grid):
    """检测迭代序列是否可进行grid"""

    if hasattr(param_grid, 'items'):
        param_grid = [param_grid]

    for p in param_grid:
        for v in p.values():
            if isinstance(v, np.ndarray) and v.ndim > 1:
                raise ValueError("Parameter array should be one-dimensional.")

            check = [isinstance(v, k) for k in (list, tuple, np.ndarray)]
            if True not in check:
                raise ValueError("Parameter values should be a list.")

            if len(v) == 0:
                raise ValueError("Parameter values should be a non-empty "
                                 "list.")


@add_process_env_sig
def grid_search_mul_process(read_cash,
                            benchmark,
                            factors,
                            choice_symbols,
                            kl_pd_manager=None):
    """
    针对输入的买入因子，卖出因子，选股因子及其它参数，进行两年历史交易回测，返回结果包装AbuScoreTuple对象
    :param read_cash: 初始化资金数(int)
    :param benchmark: 交易基准对象，AbuBenchmark实例对象
    :param factors: 买入因子，卖出因子，选股因子product 最外层tuple->dict对象->字典values->list->list对象->dict对象
            eg:
                (
                    {'buy_factors': [{'class': x1}, {'class': x2}]
                    'sell_factors': [{'class': Y1}, {'class': Y2}],
                    'stock_pickers': [{'class': Z1}, {'class': Z2}]},

                    {'buy_factors': [{'class': xx1}, {'class': xx2}]
                    'sell_factors': [{'class': Yy1}, {'class': Yy2}],
                    'stock_pickers': [{'class': Zz1}, {'class': Zz2}]
                    .................................................
                )
    :param choice_symbols: 初始备选交易对象序列
    :param kl_pd_manager: 金融时间序列管理对象，KLManager实例
    :return: AbuScoreTuple对象
    """
    # 由于grid_search_mul_process以处于多任务运行环境，所以不内部不再启动多任务，使用1个进程选股
    n_process_pick_stock = 1
    # 由于grid_search_mul_process以处于多任务运行环境，所以不内部不再启动多任务，使用1个进程择时
    n_process_pick_time = 1
    # 由于grid_search_mul_process以处于多任务运行环境，所以不内部不再启动多任务，使用1个进程数据收集
    n_process_kl = 1
    # 每一个任务子进程中返回的由AbuScoreTuple组成的独立结果对象，进程承接层使用chain.from_iterable摊开展平
    result_tuple_array = []
    # 如果因子组合的个数大于4组显示外层的进度条，否则显示内层子进程择时进程条
    show_outer_progress = True if len(factors) >= 4 else False

    with MulPidProgress(len(factors),
                        'grid search total progress',
                        show_progress=show_outer_progress) as progress:
        progress.display_step = 1
        for epoch, factor in enumerate(factors):
            progress.show(epoch + 1)
            buy_factors = factor['buy_factors']
            sell_factors = factor['sell_factors']
            stock_pickers = factor['stock_pickers']

            # 通过初始化资金数，交易基准对象构造资金管理对象capital
            capital = Capital(read_cash, benchmark)
            if stock_pickers is not None:
                # 有选股因子序列首选进行选股
                raise ("stock_pickers is not None")
                #choice_symbols = \
                #    AbuPickStockMaster.do_pick_stock_with_process(capital, benchmark,
                #                                                  stock_pickers,
                #                                                  choice_symbols=choice_symbols,
                #                                                  n_process_pick_stock=n_process_pick_stock)

            if choice_symbols is None or len(choice_symbols) == 0:
                kd_logger.info('pick stock result is zero!')
                result_tuple_array.append(
                    ScoreTuple(None, None, capital, benchmark, buy_factors,
                               sell_factors, stock_pickers))
                continue

            # 通过买入因子，卖出因子等进行择时操作
            orders_pd, action_pd, all_fit_symbols_cnt = PickTimeMaster.do_symbols_with_same_factors_process(
                choice_symbols,
                benchmark,
                buy_factors,
                sell_factors,
                capital,
                kl_pd_manager=kl_pd_manager,
                n_process_kl=n_process_kl,
                n_process_pick_time=n_process_pick_time,
                show_progress=not show_outer_progress)
            # 将最终结果包装为AbuScoreTuple对象

            result_tuple = ScoreTuple(orders_pd, action_pd, capital, benchmark,
                                      buy_factors, sell_factors, stock_pickers)
            result_tuple_array.append(result_tuple)
    return result_tuple_array


class GridSearch(object):
    """最优grid search对外接口类"""

    @classmethod
    def combine_same_factor_class(cls, factors):
        """
        合并不同的class factor到符合grid search格式的因子参数组合：

        eg：
            org_factor = [ {
                  'class': AClass,
                  'xd' : 20,
                  'past_factor': 2,
                  'up_deg_threshold': 3
             }, {
                  'class': AClass,
                  'xd' : 30,
                  'past_factor': 3,
                  'up_deg_threshold': 4
             }, {
                  'class': BClass,
                  'xd' : 20,
                  'past_factor': 2,
                  'down_deg_threshold': -2
            }, {
                  'class': BClass,
                  'xd' : 30,
                  'past_factor': 3,
                  'down_deg_threshold': -4
            }]

            转换合并后结果：

            [{'class': [AClass],
              'down_deg_threshold': [-4, -2],
              'past_factor': [2, 3],
              'xd': [20, 30]},
             {'class': [BClass],
              'past_factor': [2, 3],
              'up_deg_threshold': [3, 4],
              'xd': [20, 30]}]
        :param factors: 转换前多个买入或者卖出策略因子组成的list容器对象
        :return: 转换后符合grid search格式的策略因子组成的list容器对象
        """
        #  转换后符合grid search格式的策略因子组成的list容器对象, 首先筛选出独立参数策略
        combine_factor_list = list(
            filter(lambda factor: isinstance(factor['class'], list), factors))

        # 把需要再次组合参数的进行筛选
        factors = list(
            filter(lambda factor: not isinstance(factor['class'], list),
                   factors))

        # 先找出唯一的class集合序列
        # noinspection PyTypeChecker
        unique_class_set = set([factor['class'] for factor in factors])

        for class_value in unique_class_set:
            # 一个一个唯一的class筛选出来
            unique_class_factors = list(
                filter(lambda factor: factor['class'] == class_value, factors))
            # 每一个唯一的class筛选出来所有的字典key
            all_keys = set([
                factor_key for factor in unique_class_factors
                for factor_key in factor.keys()
            ])
            combine_factor = dict()
            for factor_key in all_keys:
                # 将有相同key的组成一个序列
                factor_grid_list = set([
                    factor[factor_key] for factor in unique_class_factors
                    if factor_key in factor.keys()
                ])
                # 新的dict符合grid search标准
                combine_factor[factor_key] = list(factor_grid_list)
            combine_factor_list.append(combine_factor)
        return combine_factor_list

    @classmethod
    def create_top_constraints_metrics(cls,
                                       constraints,
                                       scores,
                                       score_tuple_array,
                                       top_cnt=10):
        direction = int(top_cnt / abs(top_cnt))
        # 需要使用-direction因为分数是从小－>大排序的
        scores = scores[::-direction]
        result_top = constraints(scores, score_tuple_array, top_cnt)

        for top_tuple in result_top:
            kd_logger.info(u'买入策略:{}'.format(top_tuple.buy_factors))
            kd_logger.info(u'卖出策略:{}'.format(top_tuple.sell_factors))
            MetricsBase.show_general(top_tuple.orders_pd,
                                     top_tuple.action_pd,
                                     top_tuple.capital,
                                     top_tuple.benchmark,
                                     returns_cmp=True,
                                     only_info=True)
            kd_logger.info('\n')

    @classmethod
    def create_top_score_metrics(cls, scores, score_tuple_array, top_cnt=10):
        """
        类方法，根据grid_search寻找最优参数的grid结果返回的scores序列与AbuScoreTuple对象序列
        显示top n个度量结果
        :param scores: grid_search结果返回的scores序列
        :param score_tuple_array: grid_search结果返回的AbuScoreTuple对象序列
        :param top_cnt: 显示的top度量结果个数，默认10个， 如果传递的top_cnt为负数，eg：-10 则取出最不好的10个结
        """
        # 如果传递的top_cnt为负数，eg：-10 则取出最不好的10个结果
        direction = int(top_cnt / abs(top_cnt))
        # 需要使用-direction因为分数是从小－>大排序的
        top_n = scores[::-direction][:abs(top_cnt)]
        for top in top_n.index:
            top_tuple = score_tuple_array[top]
            kd_logger.info(u'买入策略:{}'.format(top_tuple.buy_factors))
            kd_logger.info(u'卖出策略:{}'.format(top_tuple.sell_factors))
            MetricsBase.show_general(top_tuple.orders_pd,
                                     top_tuple.action_pd,
                                     top_tuple.capital,
                                     top_tuple.benchmark,
                                     returns_cmp=True,
                                     only_info=True)
            kd_logger.info('\n')

    @classmethod
    def grid_search(cls,
                    choice_symbols,
                    buy_factors,
                    sell_factors,
                    read_cash=10000000,
                    score_weights=None,
                    metrics_class=None):
        """
        类方法: 不gird选股因子，只使用买入因子和卖出因子序列的gird product行为
        :param choice_symbols: 初始备选交易对象序列
        :param buy_factors: 买入因子grid序列或者直接为独立买入因子grid
        :param sell_factors: 卖出因子grid序列或者直接为独立卖出因子grid
        :param read_cash: 初始化资金数(int), 默认10000000
        :param score_weights: make_scorer中设置的评分权重
        :param metrics_class: make_scorer中设置的度量类
        """
        scores = None
        score_tuple_array = None
        if True:

            def factor_grid(factors):
                if isinstance(factors, dict):
                    # 独立因子grid, 确保参数为序列，且形成独立grid list
                    factors_grid = [{
                        bf_key: factors[bf_key] if isinstance(
                            factors[bf_key], list) else [factors[bf_key]]
                        for bf_key in factors.keys()
                    }]
                elif isinstance(factors, list):
                    # 如果传递进来的本身就是序列，需要对序列内容做参数监测，
                    factors_grid = []
                    for factor in factors:
                        factor_dict = {
                            bf_key: factor[bf_key] if isinstance(
                                factor[bf_key], list) else [factor[bf_key]]
                            for bf_key in factor.keys()
                        }
                        factors_grid.append(factor_dict)
                else:
                    raise TypeError(
                        "factors must be dict or list not {}".format(
                            type(factors)))
                return factor_dict

            # print('buy_factors', buy_factors)
            buy_factors_grid = factor_grid(buy_factors)
            # print('buy_factors_grid', buy_factors_grid)
            sell_factors_grid = factor_grid(sell_factors)

            from ultron.ump.metrics.grid_helper import gen_factor_grid, K_GEN_FACTOR_PARAMS_BUY, K_GEN_FACTOR_PARAMS_SELL

            buy_factors_product = gen_factor_grid(K_GEN_FACTOR_PARAMS_BUY,
                                                  buy_factors_grid)
            sell_factors_product = gen_factor_grid(K_GEN_FACTOR_PARAMS_SELL,
                                                   sell_factors_grid)

            kd_logger.info(u'卖出因子参数共有{}种组合方式'.format(
                len(sell_factors_product)))
            kd_logger.info(u'卖出因子组合0: 形式为{}'.format(sell_factors_product[0]))

            kd_logger.info(u'买入因子参数共有{}种组合方式'.format(len(buy_factors_product)))
            kd_logger.info(u'买入因子组合0: 形式为{}'.format(buy_factors_product[0]))

            kd_logger.info(u'买入卖出组合形式共: {} X {} = {}种'.format(
                len(buy_factors_product), len(sell_factors_product),
                len(buy_factors_product) * len(sell_factors_product)))

            gs = cls(read_cash,
                     choice_symbols,
                     buy_factors_product=buy_factors_product,
                     sell_factors_product=sell_factors_product,
                     score_weights=score_weights,
                     metrics_class=metrics_class)

            scores, score_tuple_array = gs.fit(n_jobs=-1)
            best_score_tuple_grid = gs.best_score_tuple_grid

            kd_logger.info(u'最佳买入因子参数组合：{}'.format(
                best_score_tuple_grid.buy_factors))
            kd_logger.info(u'最佳卖出因子参数组合：{}'.format(
                best_score_tuple_grid.sell_factors))
            kd_logger.info('*' * 100)

            MetricsBase.show_general(best_score_tuple_grid.orders_pd,
                                     best_score_tuple_grid.action_pd,
                                     best_score_tuple_grid.capital,
                                     best_score_tuple_grid.benchmark,
                                     returns_cmp=True,
                                     only_info=True)

        return scores, score_tuple_array

    def __init__(self,
                 read_cash,
                 choice_symbols,
                 stock_pickers_product=None,
                 buy_factors_product=None,
                 sell_factors_product=None,
                 score_weights=None,
                 metrics_class=None,
                 benchmark_kl_pd=None):
        """
        :param read_cash: 初始化资金数(int)
        :param choice_symbols: 初始备选交易对象序列
        :param stock_pickers_product: 选股因子product之后的序列
        :param buy_factors_product: 买入因子product之后的序列
        :param sell_factors_product: 卖出因子product之后的序列
        :param score_weights: make_scorer中设置的评分权重
        :param metrics_class: make_scorer中设置的度量类
        """
        self.read_cash = read_cash
        self.benchmark = Benchmark(benchmark_kl_pd=benchmark_kl_pd)
        self.kl_pd_manager = KLManager(self.benchmark,
                                       Capital(self.read_cash, self.benchmark))
        self.choice_symbols = choice_symbols
        self.stock_pickers_product = [
            None
        ] if stock_pickers_product is None else stock_pickers_product
        self.buy_factors_product = [
            None
        ] if buy_factors_product is None else buy_factors_product
        self.sell_factors_product = [
            None
        ] if sell_factors_product is None else sell_factors_product
        self.score_weights = score_weights
        self.metrics_class = metrics_class

    def fit(self, score_class=WrsmScorer, n_jobs=-1):
        """
        开始寻找最优因子参数组合，费时操作，迭代所有因子组合进行交易回测，回测结果进行评分
        :param score_class: 对回测结果进行评分的评分类，AbuBaseScorer类型，非对象，只传递类信息
        :param n_jobs: 默认回测并行的任务数，默认-1, 即启动与cpu数量相同的进程数
        :return: (scores: 评分结果dict， score_tuple_array: 因子组合序列)
        """
        pass_kl_pd_manager = None
        if len(self.stock_pickers_product
               ) == 1 and self.stock_pickers_product[0] is None:
            # 如果没有设置选股因子，外层统一进行交易数据收集，之所以是1，以为在__init__中[None]的设置
            #need_batch_gen = self.kl_pd_manager.filter_pick_time_choice_symbols(
            #    self.choice_symbols)
            # grid多进程symbol数量大于40才使用多进程，否则单进程执行
            #self.kl_pd_manager.batch_get_pick_time_kl_pd(
            #    need_batch_gen,
            #    n_process=env.g_cpu_cnt if len(need_batch_gen) > 40 else 1)
            pass_kl_pd_manager = self.kl_pd_manager

        if n_jobs <= 0:
            # 因为下面要根据n_jobs来split_k_market
            n_jobs = env.g_cpu_cnt

        factors_product = [{
            'buy_factors': item[0],
            'sell_factors': item[1],
            'stock_pickers': item[2]
        } for item in product(self.buy_factors_product, self.
                              sell_factors_product, self.stock_pickers_product)
                           ]

        # 将factors切割为n_jobs个子序列，这样可以每个进程处理一个子序列
        process_factors = split_k_market(n_jobs,
                                         market_symbols=factors_product)
        # 因为切割会有余数，所以将原始设置的进程数切换为分割好的个数, 即32 -> 33 16 -> 17
        n_jobs = len(process_factors)
        parallel = Parallel(n_jobs=n_jobs, verbose=0, pre_dispatch='2*n_jobs')
        # 多任务环境下的内存环境拷贝对象AbuEnvProcess
        p_env = EnvProcess()
        # 多层迭代各种类型因子，没一种因子组合作为参数启动一个新进程，运行grid_search_mul_process
        out_abu_score_tuple = parallel(
            delayed(grid_search_mul_process)(self.read_cash,
                                             self.benchmark,
                                             factors,
                                             self.choice_symbols,
                                             pass_kl_pd_manager,
                                             env=p_env)
            for factors in process_factors)

        # 都完事时检测一下还有没有ui进度条
        #Progress.do_check_process_is_dead()
        # 返回的AbuScoreTuple序列转换score_tuple_array, 即摊开多个子结果序列eg: ([], [], [], [])->[]
        score_tuple_array = list(chain.from_iterable(out_abu_score_tuple))
        # 使用ABuMetricsScore中make_scorer对多个参数组合的交易结果进行评分，详情阅读ABuMetricsScore模块
        scores = make_scorer(score_tuple_array,
                             score_class,
                             weights=self.score_weights,
                             metrics_class=self.metrics_class)
        # 评分结果最好的赋予best_score_tuple_grid
        self.best_score_tuple_grid = score_tuple_array[scores.index[-1]]
        return scores, score_tuple_array



# -*- encoding:utf-8 -*-
"""
    Deprecated警告模块
"""

import warnings

from ultron.ump.core.fixes import six


class Deprecated(object):
    """支持装饰类或者方法，在使用类或者方法时警告Deprecated信息"""

    def __init__(self, tip_info=''):
        # 用户自定义警告信息tip_info
        self.tip_info = tip_info

    def __call__(self, obj):
        if isinstance(obj, six.class_types):
            # 针对类装饰
            return self._decorate_class(obj)
        else:
            # 针对方法装饰
            return self._decorate_fun(obj)

    def _decorate_class(self, cls):
        """实现类装饰警告Deprecated信息"""

        msg = "class {} is deprecated".format(cls.__name__)
        if self.tip_info:
            msg += "; {}".format(self.tip_info)
        # 取出原始init
        init = cls.__init__

        def wrapped(*args, **kwargs):
            warnings.warn(msg, category=DeprecationWarning)
            return init(*args, **kwargs)

        cls.__init__ = wrapped

        wrapped.__name__ = '__init__'
        wrapped.__doc__ = self._update_doc(init.__doc__)
        # init成为deprecated_original，必须要使用这个属性名字，在其它地方，如AbuParamBase会寻找原始方法找它
        wrapped.deprecated_original = init

        return cls

    def _decorate_fun(self, fun):
        """实现方法装饰警告Deprecated信息"""

        msg = "func {} is deprecated".format(fun.__name__)
        if self.tip_info:
            msg += "; {}".format(self.tip_info)

        def wrapped(*args, **kwargs):
            warnings.warn(msg, category=DeprecationWarning)
            return fun(*args, **kwargs)

        # 更新func及文档信息
        wrapped.__name__ = fun.__name__
        wrapped.__dict__ = fun.__dict__
        wrapped.__doc__ = self._update_doc(fun.__doc__)

        return wrapped

    def _update_doc(self, func_doc):
        """更新文档信息，把原来的文档信息进行合并格式化, 即第一行为deprecated_doc(Deprecated: tip_info)，下一行为原始func_doc"""
        deprecated_doc = "Deprecated"
        if self.tip_info:
            """如果有tip format tip"""
            deprecated_doc = "{}: {}".format(deprecated_doc, self.tip_info)
        if func_doc:
            # 把原来的文档信息进行合并格式化, 即第一行为deprecated_doc，下一行为原始func_doc
            func_doc = "{}\n{}".format(deprecated_doc, func_doc)
        return func_doc



# -*- coding: utf-8 -*-

from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
import pandas as pd

from ultron.factor.data.processing import factor_processing
from ultron.factor.data.standardize import standardize
from ultron.factor.data.winsorize import winsorize_normal
from ultron.optimize.constraints import Constraints
from ultron.optimize.constraints import LinearConstraints
from ultron.optimize.linearbuilder import linear_builder
from ultron.optimize.rankbuilder import rank_builder
from ultron.optimize.longshortbulder import long_short_builder
from ultron.optimize.meanvariancebuilder import mean_variance_builder
from ultron.optimize.meanvariancebuilder import target_vol_builder
from ultron.factor.analysis.simplesettle import simple_settle


def factor_analysis(factors: pd.DataFrame,
                    factor_weights: np.ndarray,
                    industry: np.ndarray,
                    d1returns: np.ndarray = None,
                    detail_analysis=True,
                    benchmark: Optional[np.ndarray] = None,
                    risk_exp: Optional[np.ndarray] = None,
                    is_tradable: Optional[np.ndarray] = None,
                    constraints: Optional[Constraints] = None,
                    method='risk_neutral',
                    **kwargs) -> Tuple[pd.DataFrame, Optional[pd.DataFrame]]:
    if 'pre_process' in kwargs:
        pre_process = kwargs['pre_process']
        del kwargs['pre_process']
    else:
        pre_process = [winsorize_normal, standardize]

    if 'post_process' in kwargs:
        post_process = kwargs['post_process']
        del kwargs['post_process']
    else:
        post_process = [winsorize_normal, standardize]

    er = factor_processing(factors.values, pre_process, risk_exp,
                           post_process) @ factor_weights

    return er_portfolio_analysis(er, industry, d1returns, constraints,
                                 detail_analysis, benchmark, is_tradable,
                                 method, **kwargs)


def er_portfolio_analysis(
        er: np.ndarray,
        industry: np.ndarray,
        dx_return: np.ndarray,
        constraints: Optional[Union[LinearConstraints, Constraints]] = None,
        detail_analysis=True,
        benchmark: Optional[np.ndarray] = None,
        is_tradable: Optional[np.ndarray] = None,
        method='risk_neutral',
        **kwargs) -> Tuple[pd.DataFrame, Optional[pd.DataFrame]]:
    er = er.flatten()

    def create_constraints(benchmark, **kwargs):
        if 'lbound' in kwargs:
            lbound = kwargs['lbound'].copy()
            del kwargs['lbound']
        else:
            lbound = np.maximum(0., benchmark - 0.01)

        if 'ubound' in kwargs:
            ubound = kwargs['ubound'].copy()
            del kwargs['ubound']
        else:
            ubound = 0.01 + benchmark
        if is_tradable is not None:
            ubound[~is_tradable] = np.minimum(lbound, ubound)[~is_tradable]

        risk_lbound, risk_ubound = constraints.risk_targets()
        cons_exp = constraints.risk_exp
        return lbound, ubound, cons_exp, risk_lbound, risk_ubound

    if method == 'risk_neutral' or method == 'long_risk_neutral':
        lbound, ubound, cons_exp, risk_lbound, risk_ubound = create_constraints(
            benchmark, **kwargs)
        turn_over_target = kwargs.get('turn_over_target')
        current_position = kwargs.get('current_position')
        status, _, weights = linear_builder(er,
                                            risk_constraints=cons_exp,
                                            lbound=lbound,
                                            ubound=ubound,
                                            risk_target=(risk_lbound,
                                                         risk_ubound),
                                            turn_over_target=turn_over_target,
                                            current_position=current_position)
        if status not in ("optimal", "optimal_inaccurate"):
            raise ValueError(
                'linear programming optimizer in status: {0}'.format(status))

    elif method == 'rank':
        weights = rank_builder(
            er, use_rank=kwargs['use_rank'],
            masks=is_tradable).flatten() * benchmark.sum() / kwargs['use_rank']
    elif method == 'ls' or method == 'long_short':
        weights = long_short_builder(er).flatten()
    elif method == 'mv' or method == 'mean_variance' or method == 'long_mean_variance':
        lbound, ubound, cons_exp, risk_lbound, risk_ubound = create_constraints(
            benchmark, **kwargs)
        risk_model = kwargs['risk_model']

        if 'lam' in kwargs:
            lam = kwargs['lam']
        else:
            lam = 1.

        status, _, weights = mean_variance_builder(er,
                                                   risk_model=risk_model,
                                                   bm=benchmark,
                                                   lbound=lbound,
                                                   ubound=ubound,
                                                   risk_exposure=cons_exp,
                                                   risk_target=(risk_lbound,
                                                                risk_ubound),
                                                   lam=lam)
        if status != 'optimal':
            raise ValueError(
                'mean variance optimizer in status: {0}'.format(status))

    elif method == 'tv' or method == 'target_vol':
        lbound, ubound, cons_exp, risk_lbound, risk_ubound = create_constraints(
            benchmark, **kwargs)
        risk_model = kwargs['risk_model']

        if 'target_vol' in kwargs:
            target_vol = kwargs['target_vol']
        else:
            target_vol = 1.

        status, _, weights = target_vol_builder(er,
                                                risk_model=risk_model,
                                                bm=benchmark,
                                                lbound=lbound,
                                                ubound=ubound,
                                                risk_exposure=cons_exp,
                                                risk_target=(risk_lbound,
                                                             risk_ubound),
                                                vol_target=target_vol)
    else:
        raise ValueError("Unknown building type ({0})".format(method))

    if detail_analysis:
        analysis = simple_settle(weights, dx_return, industry, benchmark)
    else:
        analysis = None
    return pd.DataFrame({'weight': weights,
                         'industry': industry,
                         'er': er}), \
           analysis



# -*- encoding:utf-8 -*-
"""
    风险控制仓位管理基础
"""
from abc import ABCMeta, abstractmethod
from ultron.ump.core.fixes import six
from ultron.ump.market.market import MarketMixin
"""每一笔交易最大仓位比例设置，外部可通过如：beta.position.g_pos_max = 0.5修改最大每一笔交易最大仓位比例，默认75%"""
g_pos_max = 0.75
"""
    保证金最小比例，默认1，即不使用融资，不会触发Margin Call
    在期货数据中有每种商品最少保证金比例，可使用设置
    外部可通过如：abupy.beta.position.g_deposit_rate = 0.5
"""
g_deposit_rate = 1
"""
    买入因子全局默认仓位管理类，默认None的情况下会使用AbuAtrPosition作为默认仓位管理类.

    和卖出因子，选股因子不同，一个买入因子可以对应多个卖出因子，多个选股因子，但一个买入
    因子只能对应一个仓位管理类，可以是全局仓位管理，也可以是针对买入因子的独有附属仓位管理
    类
"""
g_default_pos_class = None


class PositionBase(six.with_metaclass(ABCMeta, MarketMixin)):
    """仓位管理抽象基类"""

    def __init__(self, kl_pd_buy, factor_name, symbol_name, bp, read_cash,
                 **kwargs):
        """
        :param kl_pd_buy: 交易当日的交易数据
        :param factor_name: 因子名称
        :param symbol_name: symbol代码
        :param bp: 买入价格
        :param read_cash: 初始资金
        :param deposit_rate: 保证金比例
        """
        self.kl_pd_buy = kl_pd_buy
        self.factor_name = factor_name
        self.symbol_name = symbol_name
        self.bp = bp
        self.read_cash = read_cash

        # 如果有全局最大仓位设置基类负责弹出
        self.pos_max = kwargs.pop('pos_max', g_pos_max)
        # 如果有全局保证金最小比例设置基类负责弹出
        self.deposit_rate = kwargs.pop('deposit_rate', g_deposit_rate)

        # 子类继续完成自有的构造
        self._init_self(**kwargs)

    def __str__(self):
        """打印对象显示：class name, factor_name, symbol_name, read_cash, deposit_rate"""
        return '{}: factor_name:{}, symbol_name:{}, read_cash:{}, deposit_rate:{}'.format(
            self.__class__.__name__, self.factor_name, self.symbol_name,
            self.read_cash, self.deposit_rate)

    __repr__ = __str__

    @abstractmethod
    def _init_self(self, **kwargs):
        """子类仓位管理针对可扩展参数的初始化"""
        pass

    @abstractmethod
    def fit_position(self, factor_object):
        """
        fit_position计算的结果是买入多少个单位（股，手，顿，合约）具体计算子类实现
        :param factor_object: ABuFactorBuyBases实例对象
        :return:买入多少个单位（股，手，顿，合约）
        """
        pass


# -*- coding: utf-8 -*-
from asyncio import constants
from typing import Union
import cvxpy as cp
import numpy as np


class TargetVarianceOptimizer(object):

    def __init__(self,
                 cost: np.ndarray,
                 benchmark: np.ndarray,
                 l1norm: float,
                 variance_target: float,
                 variance: np.ndarray = None,
                 lower_bound: Union[float, np.ndarray] = None,
                 upper_bound: Union[float, np.ndarray] = None):
        self._n = len(cost)
        self._lower_bound = lower_bound
        self._upper_bound = upper_bound
        self._l1norm = l1norm
        self._variance = variance
        self._variance_target = variance_target
        self._benchmark = benchmark
        self._cost = cost

    def _prepare(self):
        x = cp.Variable(self._n)
        constraints = []
        if self._lower_bound is not None:
            constraints.append(x >= self._lower_bound)
        if self._upper_bound is not None:
            constraints.append(x <= self._upper_bound)
        return x, constraints

    def solver(self, solver: str = "ECOS"):
        x, constraints = self._prepare()
        risk = cp.quad_form(x, self._variance)
        constraints.append(cp.pnorm(x - self._benchmark, 1) <= self._l1norm)
        constraints.append(cp.pnorm(x, 1) <= 2.0)
        constraints.append(risk <= self._variance_target)

        constraints.append(sum(x) == 0.0)

        prob = cp.Problem(cp.Minimize(x @ self._cost), constraints=constraints)
        prob.solve(solver=solver)
        return x.value, prob.value, prob.status



# -*- coding: utf-8 -*-

import unittest

import numpy as np
from sklearn.linear_model import LinearRegression

from ultron.factor.data.neutralize import neutralize


class Neutralize(unittest.TestCase):

    def setUp(self):
        self.y = np.random.randn(3000, 4)
        self.x = np.random.randn(3000, 10)
        self.groups = np.random.randint(30, size=3000)

    def test_neutralize(self):
        calc_res = neutralize(self.x, self.y)

        model = LinearRegression(fit_intercept=False)
        model.fit(self.x, self.y)

        exp_res = self.y - self.x @ model.coef_.T

        np.testing.assert_array_almost_equal(calc_res, exp_res)

    def test_neutralize_with_group(self):

        calc_res = neutralize(self.x, self.y, self.groups)

        model = LinearRegression(fit_intercept=False)
        for i in range(30):
            curr_x = self.x[self.groups == i]
            curr_y = self.y[self.groups == i]
            model.fit(curr_x, curr_y)
            exp_res = curr_y - curr_x @ model.coef_.T
            np.testing.assert_array_almost_equal(calc_res[self.groups == i], exp_res)

    def test_neutralize_explain_output(self):
        y = self.y[:, 0].flatten()

        calc_res, other_stats = neutralize(self.x, y, detail=True)

        model = LinearRegression(fit_intercept=False)
        model.fit(self.x, y)

        exp_res = y - self.x @ model.coef_.T
        exp_explained = self.x * model.coef_.T

        np.testing.assert_array_almost_equal(calc_res, exp_res.reshape(-1, 1))
        np.testing.assert_array_almost_equal(other_stats['explained'][:, :, 0], exp_explained)

        calc_res, other_stats = neutralize(self.x, self.y, detail=True)

        model = LinearRegression(fit_intercept=False)
        model.fit(self.x, self.y)

        exp_res = self.y - self.x @ model.coef_.T
        np.testing.assert_array_almost_equal(calc_res, exp_res)

        for i in range(self.y.shape[1]):
            exp_explained = self.x * model.coef_.T[:, i]
            np.testing.assert_array_almost_equal(other_stats['explained'][:, :, i], exp_explained)

    def test_neutralize_explain_output_with_group(self):
        y = self.y[:, 0].flatten()

        calc_res, other_stats = neutralize(self.x, y, self.groups, detail=True)

        model = LinearRegression(fit_intercept=False)
        for i in range(30):
            curr_x = self.x[self.groups == i]
            curr_y = y[self.groups == i]
            model.fit(curr_x, curr_y)
            exp_res = curr_y - curr_x @ model.coef_.T
            exp_explained = curr_x * model.coef_.T
            np.testing.assert_array_almost_equal(calc_res[self.groups == i], exp_res.reshape(-1, 1))
            np.testing.assert_array_almost_equal(other_stats['explained'][self.groups == i, :, 0],
                                                 exp_explained)

        calc_res, other_stats = neutralize(self.x, self.y, self.groups, detail=True)

        model = LinearRegression(fit_intercept=False)
        for i in range(30):
            curr_x = self.x[self.groups == i]
            curr_y = self.y[self.groups == i]
            model.fit(curr_x, curr_y)
            exp_res = curr_y - curr_x @ model.coef_.T
            np.testing.assert_array_almost_equal(calc_res[self.groups == i], exp_res)

            for j in range(self.y.shape[1]):
                exp_explained = curr_x * model.coef_.T[:, j]
                np.testing.assert_array_almost_equal(
                    other_stats['explained'][self.groups == i, :, j], exp_explained)


# -*- encoding:utf-8 -*-
"""
    市场相关切割，选股，等操作模块
"""
import pandas as pd
from ultron.kdutils.lazy import LazyFunc
from ultron.ump.core.env import EMarketTargetType
from ultron.ump.market.symbol import Symbol, code_to_symbol
from ultron.ump.market.stock import SymbolCN
from ultron.ump.market.futures import FuturesCn


class MarketMixin(object):
    """
        市场信息混入类，被混入类需要设置self.symbol_name，
        通过code_to_symbol将symbol转换为Symbol对象, 通过Symbol对象
        查询market和sub_market
    """

    @LazyFunc
    def _symbol(self):
        """通过code_to_symbol将symbol转换为Symbol对象 LazyFunc"""
        if not hasattr(self, 'symbol_name'):
            # 被混入类需要设置self.symbol_name
            raise NameError('must name symbol_name!!')
        # 通过code_to_symbol将symbol转换为Symbol对象
        return code_to_symbol(self.symbol_name)

    @LazyFunc
    def symbol_market(self):
        """查询self.symbol_name的市场类型 LazyFunc"""
        return self._symbol.market

    @LazyFunc
    def symbol_sub_market(self):
        """查询self.symbol_name的子市场类型，即交易所信息 LazyFunc"""
        return self._symbol.sub_market


def _all_tc_symbol():
    """
    获取币类symbol，注意这里只取比特币与莱特币，可自行扩展其它币种
    :return:
    """
    return ['btc', 'ltc']


def _all_futures_cn():
    """
    AbuFuturesCn().symbol获取国内期货symbol代码，注意这里只取连续合约代码
    :return:
    """
    return FuturesCn().symbol


def _all_cn_symbol(index=False):
    """
    通过AbuSymbolCN().all_symbol获取A股全市场股票代码
    :param index: 是否包含指数
    :return:
    """
    # noinspection PyProtectedMember
    #if ABuEnv._g_enable_example_env_ipython:
    #    return K_SAND_BOX_CN
    return SymbolCN().all_symbol(index=index)


def split_k_market(k_split, market_symbols=None, market=None):
    """
    将market_symbols序列切分成k_split个序列
    :param k_split: 切分成子序列个数int
    :param market_symbols: 待切割的原始symbols序列，如果none, 将取market参数中指定的市场所有symbol
    :param market: 默认None，如None则服从ABuEnv.g_market_target市场设置
    :return: list序列，序列中的每一个元素都是切割好的子symbol序列
    """
    if market_symbols is None:
        # 取market参数中指定的市场所有symbol
        market_symbols = all_symbol(market=market)
    if len(market_symbols) < k_split:
        # 特殊情况，eg：k_split＝100，但是len(market_symbols)＝50，就不切割了，直接返回
        return [[symbol] for symbol in market_symbols]

    # 计算每一个子序列的承载的symbol个数，即eg：100 ／ 5 ＝ 20
    sub_symbols_cnt = int(len(market_symbols) / k_split)
    group_adjacent = lambda a, k: zip(*([iter(a)] * k))
    # 使用lambda函数group_adjacent将market_symbols切割子序列，每个子系列sub_symbols_cnt个
    symbols = list(group_adjacent(market_symbols, sub_symbols_cnt))
    # 将不能整除的余数symbol个再放进去
    residue_ind = -(len(market_symbols) %
                    sub_symbols_cnt) if sub_symbols_cnt > 0 else 0
    if residue_ind < 0:
        # 所以如果不能除尽，最终切割的子序列数量为k_split+1, 外部如果需要进行多认为并行，可根据最终切割好的数量重分配任务数
        symbols.append(market_symbols[residue_ind:])
    return symbols


def all_symbol(market=None, ss=False, index=False, value=True):
    """
    根据传入的市场获取全市场代码
    :param market: 默认None，如None则服从ABuEnv.g_market_target市场设置
    :param ss: 是否将返回序列使用pd.Series包装
    :param index: 是否包含指数大盘symbol
    :param value: 返回字符串值，即如果序列中的元素是Symbol对象，Symbol转换字符串
    :return:
    """
    if market is None:
        raise
        # None则服从ABuEnv.g_market_target市场设置
        #market = ABuEnv.g_market_target

    if market == EMarketTargetType.E_MARKET_TARGET_CN:
        symbols = _all_cn_symbol(index)
    elif market == EMarketTargetType.E_MARKET_TARGET_FUTURES_CN:
        symbols = _all_futures_cn()
    elif market == EMarketTargetType.E_MARKET_TARGET_TC:
        symbols = _all_tc_symbol()
    else:
        raise TypeError('JUST SUPPORT EMarketTargetType!')

    # 在出口统一确保唯一性, 在每一个内部_all_xx_symbol中也要尽量保证唯一
    symbols = list(set(symbols))

    if value:
        """
            如果是Symbol类型的还原成字符串，尽量在上面返回的symbols序列是字符串类型
            不要在上面构造symbol，浪费效率，统一会在之后的逻辑中通过code_to_symbol
            进行转换
        """
        symbols = [
            sb.value if isinstance(sb, Symbol) else sb for sb in symbols
        ]
    # 根据参数ss是否将返回序列使用pd.Series包装
    return pd.Series(symbols) if ss else symbols


# -*- encoding:utf-8 -*-
"""
    金融时间序列管理模块
"""
import pandas as pd


class KLManager(object):
    """金融时间序列管理类"""

    def __init__(self, benchmark, capital):
        """
        :param benchmark: 交易基准对象，AbuBenchmark实例对象
        :param capital: 资金类AbuCapital实例化对象
        """
        self.benchmark = benchmark
        self.capital = capital
        # 选股时间交易序列字典
        pick_stock_kl_pd_dict = dict()
        # 择时时间交易序列字典
        pick_time_kl_pd_dict = dict()
        # 类字典pick_kl_pd_dict将选股和择时字典包起来
        self.pick_kl_pd_dict = {
            'pick_stock': pick_stock_kl_pd_dict,
            'pick_time': pick_time_kl_pd_dict
        }

    def __str__(self):
        """打印对象显示：pick_stock + pick_time keys, 即所有symbol信息"""
        keys = set(self.pick_kl_pd_dict['pick_stock'].keys()) | set(
            self.pick_kl_pd_dict['pick_time'].keys())
        return 'pick_stock + pick_time keys :{}'.format(keys)

    __repr__ = __str__

    def __len__(self):
        """对象长度：选股字典长度 + 择时字典长度"""
        return len(self.pick_kl_pd_dict['pick_stock']) + len(
            self.pick_kl_pd_dict['pick_time'])

    def __contains__(self, item):
        """成员测试：在择时字典中或者在选股字典中"""
        return item in self.pick_kl_pd_dict[
            'pick_stock'] or item in self.pick_kl_pd_dict['pick_time']

    def __missing__(self, key):
        """对象缺失：需要根据key使用code_to_symbol进行fetch数据，暂未实现"""
        # TODO 需要根据key使用code_to_symbol进行fetch数据
        raise NotImplementedError('TODO AbuKLManager __missing__')

    def __getitem__(self, key):
        """索引获取：尝试分别从选股字典，择时字典中查询，返回两个字典的查询结果"""
        pick_stock_item = None
        if key in self.pick_kl_pd_dict['pick_stock']:
            pick_stock_item = self.pick_kl_pd_dict['pick_stock'][key]
        pick_time_item = None
        if key in self.pick_kl_pd_dict['pick_time']:
            pick_time_item = self.pick_kl_pd_dict['pick_time'][key]
        return pick_stock_item, pick_time_item

    def __setitem__(self, key, value):
        """索引设置：抛错误，即不准许外部设置"""
        raise AttributeError("AbuKLManager set value!!!")

    def set_pick_time(self, values):
        self.pick_kl_pd_dict['pick_time'] = values

    def set_pick_stock(self, values):
        self.pick_kl_pd_dict['pick_stock'] = values

    def get_pick_time_kl_pd(self, target_symbol):
        """对外获取择时时段金融时间序列，首先在内部择时字典中寻找，没找到使用_fetch_pick_time_kl_pd获取，且保存择时字典"""
        if target_symbol in self.pick_kl_pd_dict['pick_time']:
            kl_pd = self.pick_kl_pd_dict['pick_time'][target_symbol]
            if kl_pd is not None:
                # 因为在多进程的时候拷贝会丢失name信息
                kl_pd.name = target_symbol
            return kl_pd
        kl_pd = pd.DataFrame()
        kl_pd.name = target_symbol
        return kl_pd


# -*- coding: utf-8 -*-

from ultron.sentry.Math.Accumulators.IAccumulators import Exp
from ultron.sentry.Math.Accumulators.IAccumulators import Log
from ultron.sentry.Math.Accumulators.IAccumulators import Sqrt
from ultron.sentry.Math.Accumulators.IAccumulators import Sign
from ultron.sentry.Math.Accumulators.IAccumulators import Abs
from ultron.sentry.Math.Accumulators.IAccumulators import Pow
from ultron.sentry.Math.Accumulators.IAccumulators import Acos
from ultron.sentry.Math.Accumulators.IAccumulators import Acosh
from ultron.sentry.Math.Accumulators.IAccumulators import Asin
from ultron.sentry.Math.Accumulators.IAccumulators import Asinh
from ultron.sentry.Math.Accumulators.IAccumulators import NormInv
from ultron.sentry.Math.Accumulators.IAccumulators import Current
from ultron.sentry.Math.Accumulators.IAccumulators import Latest
from ultron.sentry.Math.Accumulators.IAccumulators import Ceil
from ultron.sentry.Math.Accumulators.IAccumulators import Floor
from ultron.sentry.Math.Accumulators.IAccumulators import Round
from ultron.sentry.Math.Accumulators.IAccumulators import Identity
from ultron.sentry.Math.Accumulators.IAccumulators import IIF
from ultron.sentry.Math.Accumulators.IAccumulators import Sign
from ultron.sentry.Math.Accumulators.IAccumulators import Negative

from ultron.sentry.Math.Accumulators.StatelessAccumulators import Diff
from ultron.sentry.Math.Accumulators.StatelessAccumulators import SimpleReturn
from ultron.sentry.Math.Accumulators.StatelessAccumulators import LogReturn
from ultron.sentry.Math.Accumulators.StatelessAccumulators import PositivePart
from ultron.sentry.Math.Accumulators.StatelessAccumulators import NegativePart
from ultron.sentry.Math.Accumulators.StatelessAccumulators import Max
from ultron.sentry.Math.Accumulators.StatelessAccumulators import Maximum
from ultron.sentry.Math.Accumulators.StatelessAccumulators import Min
from ultron.sentry.Math.Accumulators.StatelessAccumulators import Minimum
from ultron.sentry.Math.Accumulators.StatelessAccumulators import Sum
from ultron.sentry.Math.Accumulators.StatelessAccumulators import Average
from ultron.sentry.Math.Accumulators.StatelessAccumulators import XAverage
from ultron.sentry.Math.Accumulators.StatelessAccumulators import Variance
from ultron.sentry.Math.Accumulators.StatelessAccumulators import Product

from ultron.sentry.Math.Accumulators.StatefulAccumulators import Shift
from ultron.sentry.Math.Accumulators.StatefulAccumulators import Delta
from ultron.sentry.Math.Accumulators.StatefulAccumulators import MovingAverage
from ultron.sentry.Math.Accumulators.StatefulAccumulators import MovingDecay
from ultron.sentry.Math.Accumulators.StatefulAccumulators import MovingPositiveAverage
from ultron.sentry.Math.Accumulators.StatefulAccumulators import MovingNegativeAverage
from ultron.sentry.Math.Accumulators.StatefulAccumulators import MovingPositiveDifferenceAverage
from ultron.sentry.Math.Accumulators.StatefulAccumulators import MovingNegativeDifferenceAverage
from ultron.sentry.Math.Accumulators.StatefulAccumulators import MovingRSI
from ultron.sentry.Math.Accumulators.StatefulAccumulators import MovingSum
from ultron.sentry.Math.Accumulators.StatefulAccumulators import MovingCountedPositive
from ultron.sentry.Math.Accumulators.StatefulAccumulators import MovingCountedNegative
from ultron.sentry.Math.Accumulators.StatefulAccumulators import MovingVariance
from ultron.sentry.Math.Accumulators.StatefulAccumulators import MovingStandardDeviation
from ultron.sentry.Math.Accumulators.StatefulAccumulators import MovingNegativeVariance
from ultron.sentry.Math.Accumulators.StatefulAccumulators import MovingCorrelation
from ultron.sentry.Math.Accumulators.StatefulAccumulators import MovingMax
from ultron.sentry.Math.Accumulators.StatefulAccumulators import MovingArgMax
from ultron.sentry.Math.Accumulators.StatefulAccumulators import MovingMin
from ultron.sentry.Math.Accumulators.StatefulAccumulators import MovingArgMin
from ultron.sentry.Math.Accumulators.StatefulAccumulators import MovingRank
from ultron.sentry.Math.Accumulators.StatefulAccumulators import MovingQuantile
from ultron.sentry.Math.Accumulators.StatefulAccumulators import MovingAllTrue
from ultron.sentry.Math.Accumulators.StatefulAccumulators import MovingAnyTrue
from ultron.sentry.Math.Accumulators.StatefulAccumulators import MovingProduct
from ultron.sentry.Math.Accumulators.StatefulAccumulators import MACD
from ultron.sentry.Math.Accumulators.StatefulAccumulators import MovingRank
from ultron.sentry.Math.Accumulators.StatefulAccumulators import MovingLogReturn
from ultron.sentry.Math.Accumulators.StatefulAccumulators import MovingResidue
from ultron.sentry.Math.Accumulators.StatefulAccumulators import MovingSharp
from ultron.sentry.Math.Accumulators.StatefulAccumulators import MovingSortino
from ultron.sentry.Math.Accumulators.StatefulAccumulators import MovingDrawdown
from ultron.sentry.Math.Accumulators.StatefulAccumulators import MovingMaxDrawdown


__all__ = ["Exp",
           "Log",
           "Sqrt",
           "Sign",
           "Negative",
           "Abs",
           "Pow",
           "Acos",
           "Acosh",
           "Asin",
           "Asinh",
           "NormInv",
           "Current",
           "Latest",
           "Sign",
           "Diff",
           "SimpleReturn",
           "LogReturn",
           "PositivePart",
           "NegativePart",
           "Max",
           "Maximum",
           "Min",
           "Minimum",
           "Sum",
           "Average",
           "XAverage",
           "MACD",
           "Variance",
           "Shift",
           "Delta",
           "IIF",
           "Identity",
           "MovingAverage",
           "MovingDecay",
           "MovingPositiveAverage",
           "MovingNegativeAverage",
           "MovingPositiveDifferenceAverage",
           "MovingNegativeDifferenceAverage",
           "MovingRSI",
           "MovingSum",
           "MovingCountedPositive",
           "MovingCountedNegative",
           "MovingNegativeVariance",
           "MovingCorrelation",
           "MovingMax",
           "MovingArgMax",
           "MovingMin",
           "MovingArgMin",
           "MovingRank",
           "MovingQuantile",
           "MovingAllTrue",
           "MovingAnyTrue",
           "MovingVariance",
           "MovingStandardDeviation",
           "MovingLogReturn",
           "MovingResidue",
           "MovingSharp",
           "MovingSortino",
           "MovingDrawdown",
           "MovingMaxDrawdown",
           "Product",
           "MovingProduct"]



# -*- coding: utf-8 -*-

import unittest

import numpy as np
import pandas as pd
from scipy.stats import zscore

from ultron.factor.data.standardize import Standardizer
from ultron.factor.data.standardize  import projection
from ultron.factor.data.standardize import standardize


class Standardize(unittest.TestCase):

    def setUp(self):
        self.x = np.random.randn(3000, 10)
        self.groups = np.random.randint(10, 30, size=3000)

    def test_standardize(self):
        calc_zscore = standardize(self.x)
        exp_zscore = zscore(self.x, ddof=1)

        np.testing.assert_array_almost_equal(calc_zscore, exp_zscore)

    def test_projection(self):
        calc_projected = projection(self.x)
        exp_projected = self.x / np.sqrt(np.sum(np.square(self.x), axis=1).reshape((-1, 1)))

        np.testing.assert_array_almost_equal(calc_projected, exp_projected)

    def test_projection_with_groups(self):
        calc_projected = projection(self.x, self.groups, axis=0)
        exp_projected = pd.DataFrame(self.x).groupby(
            self.groups
        ).transform(lambda s: s / np.sqrt(np.square(s).sum(axis=0)))

        np.testing.assert_array_almost_equal(calc_projected, exp_projected)

    def test_standardize_with_group(self):
        calc_zscore = standardize(self.x, self.groups)
        exp_zscore = pd.DataFrame(self.x). \
            groupby(self.groups). \
            transform(lambda s: (s - s.mean(axis=0)) / s.std(axis=0, ddof=1))
        np.testing.assert_array_almost_equal(calc_zscore, exp_zscore)

    def test_standardizer(self):
        s = Standardizer()
        s.fit(self.x)
        calc_zscore = s.transform(self.x)

        exp_zscore = standardize(self.x)
        np.testing.assert_array_almost_equal(calc_zscore, exp_zscore)
        np.testing.assert_array_almost_equal(s(self.x), exp_zscore)

    def test_grouped_standardizer(self):
        s = Standardizer()
        s.fit(self.x, self.groups)
        calc_zscore = s.transform(self.x, self.groups)

        exp_zscore = standardize(self.x, self.groups)
        np.testing.assert_array_almost_equal(calc_zscore, exp_zscore)
        np.testing.assert_array_almost_equal(s(self.x, self.groups), exp_zscore)